import kaa_testutils
from kaa import highlight
from kaa.filetype.html import htmlmode
from kaa.filetype.html import get_encoding

class TestHTMLHighlight(kaa_testutils._TestDocBase):
    tokenizers = htmlmode.build_tokenizers()

    def test_entity(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('&lt; &#1111; &#x2222;')

        assert [
            (0, 4, self.tokenizers[0].tokens.keywords.tokenid),
            (4, 5, hl.tokenizers[0].nulltoken),
            (5, 12, self.tokenizers[0].tokens.keywords.tokenid),
            (12, 13, hl.tokenizers[0].nulltoken),
            (13, 21, self.tokenizers[0].tokens.keywords.tokenid)
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_comment(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('<!--abc-->')

        assert [
            (0, 4, self.tokenizers[0].tokens.comment.span_start),
            (4, 7, self.tokenizers[0].tokens.comment.span_mid),
            (7, 10, self.tokenizers[0].tokens.comment.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('<!--abc-x->')

        assert [
            (0, 4, self.tokenizers[0].tokens.comment.span_start),
            (4, 11, self.tokenizers[0].tokens.comment.span_mid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_endtag(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('</abc>')

        assert [
            (0, 2, self.tokenizers[0].tokens.endtag.span_start),
            (2, 5, self.tokenizers[0].tokens.endtag.span_mid),
            (5, 6, self.tokenizers[0].tokens.endtag.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_elem(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('<a> ')
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 2, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (2, 3, hl.tokenizers[0].tokens.htmltag.span_gt),
            (3, 4, hl.tokenizers[0].nulltoken),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('< abc > ')
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 5, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (5, 7, hl.tokenizers[0].tokens.htmltag.span_gt),
            (7, 8, hl.tokenizers[0].nulltoken),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('<abc xyz>')
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 4, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (4, 5, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (5, 8, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (8, 9, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('<abc xyz = opq>')
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 4, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (4, 5, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (5, 8, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (8, 11, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (11, 14, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (14, 15, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('<abc xyz = "opq">')
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 4, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (4, 5, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (5, 8, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (8, 11, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (11, 16, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (16, 17, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc("<abc xyz = 'opq' abc=123>")
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 4, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (4, 5, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (5, 8, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (8, 11, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (11, 16, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (16, 17, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (17, 20, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (20, 21, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (21, 24, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (24, 25, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc("<abc xyz= ")
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 4, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (4, 5, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (5, 8, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (8, 10, hl.tokenizers[0].tokens.htmltag.span_elemws),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc("<abc xyz=>")
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 4, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (4, 5, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (5, 8, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (8, 9, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (9, 10, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_javascript(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc("&nbsp;<script>if</script>if&nbsp;")
        assert [
            (0, 6, hl.tokenizers[0].tokens.keywords.tokenid),
            (6, 14, hl.tokenizers[0].tokens.jsstart.section_start),
            (14, 16, hl.tokenizers[1].tokens[0].tokenid),
            (16, 25, hl.tokenizers[1].tokens[-1].section_end),
            (25, 27, hl.tokenizers[0].nulltoken),
            (27, 33, hl.tokenizers[0].tokens.keywords.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_style(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc("<style>a</style>&nbsp;")
        assert [
            (0, 7, hl.tokenizers[0].tokens.cssstart.section_start),
            (7, 8, hl.tokenizers[2].tokens[0].span_selector),
            (8, 16, hl.tokenizers[2].tokens[-1].span_close),
            (16, 22, hl.tokenizers[0].tokens.keywords.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_javascriptattr1(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc("<a ona='if'>")
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 2, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (2, 3, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (3, 6, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (6, 7, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (7, 8, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (8, 10, hl.tokenizers[0]
             .tokens.jsattr1.subtokenizer.tokens.keywords.tokenid),
            (10, 11, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (11, 12, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_javascriptattr2(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('<a ona="if">')
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 2, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (2, 3, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (3, 6, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (6, 7, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (7, 8, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (8, 10, hl.tokenizers[0]
             .tokens.jsattr2.subtokenizer.tokens.keywords.tokenid),
            (10, 11, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (11, 12, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_javascriptattr3(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('<a ona="if>')
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 2, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (2, 3, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (3, 6, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (6, 7, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (7, 8, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (8, 10, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (10, 11, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_javascriptattr4(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('''<a ona='//\n\'''')

        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 2, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (2, 3, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (3, 6, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (6, 7, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (7, 8, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (8, 10, hl.tokenizers[0].
                     tokens.jsattr1.subtokenizer.tokens.comment2.span_start),
            (10, 11, hl.tokenizers[0].tokens.jsattr1.subtokenizer.nulltoken),
            (11, 12, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_cssattr(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc("<a style=''>")
        assert [
            (0, 1, hl.tokenizers[0].tokens.htmltag.span_lt),
            (1, 2, hl.tokenizers[0].tokens.htmltag.span_elemname),
            (2, 3, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (3, 8, hl.tokenizers[0].tokens.htmltag.span_attrname),
            (8, 9, hl.tokenizers[0].tokens.htmltag.span_elemws),
            (9, 10, hl.tokenizers[0].tokens.htmltag.span_attrvalue),
            (10, 11, hl.tokenizers[0]
             .tokens.cssattr1.subtokenizer.tokens[0].span_close),
            (11, 12, hl.tokenizers[0].tokens.htmltag.span_gt),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_xhtml(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc("""<?xml version="1.0" encoding="utf-8" ?>""")
        assert [
            (0, 2, self.tokenizers[0].tokens.xmlpi.span_start),
            (2, 37, self.tokenizers[0].tokens.xmlpi.span_mid),
            (37, 39, self.tokenizers[0].tokens.xmlpi.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc("""<!DOCTYPE html PUBLIC>""")

        assert [
            (0, 2, self.tokenizers[0].tokens.xmldef.span_start),
            (2, 21, self.tokenizers[0].tokens.xmldef.span_mid),
            (21, 22, self.tokenizers[0].tokens.xmldef.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))


class TestEncDecl:
    def test_html5(self):
        assert b'UTF-8' == get_encoding(b'<meta charset="UTF-8">')
        assert b'UTF-8' == get_encoding(b"<meta charset='UTF-8'>")
        assert not  get_encoding(b"<meta charset='UTF-8>")

    def test_html(self):
        assert b'UTF-8' == get_encoding(
            b'''<meta http-equiv="Content-type"
            content="text/html;charset=UTF-8">''')

        assert not get_encoding(
            b'''<meta con="text/html;charset=UTF-8">''')

        assert not get_encoding(
            b'''<meta content="text/html;char=UTF-8">''')

    def test_xml(self):
        assert b'UTF-8' == get_encoding(
                b'<?xml version="1.0" encoding="UTF-8"?>')

        assert not get_encoding(
                b'<?xm version="1.0" encoding="UTF-8"?>')

        assert not get_encoding(
                b'<?xml version="1.0" encoding=""?>')



from kaa.theme import Theme, Style, Overlay

DefaultThemes = {
    'basic': [
        Style('default', 'default', 'default'),
        Style('lineno', 'White', 'Blue'),
        Style('parenthesis_cur', 'White', 'Blue'),
        Style('parenthesis_match', 'Red', 'Yellow'),

        Style('keyword', 'Magenta', 'default'),
        Style('constant', 'Red', 'default'),
        Style('directive', 'Orange', 'default'),
        Style('comment', 'Cyan', 'default'),
        Style('string', 'Blue', 'default'),
        Style('number', 'Green', 'default'),

        Overlay('cursor_row', None, 'Base02'),
        Overlay('breakpoint', None, 'Base02'),
        Overlay('current_row', None, 'Yellow'),
    ],
}



from collections import namedtuple
from kaa.filetype.default import defaultmode, theme
from kaa.highlight import Tokenizer, Keywords, Span, SingleToken
from kaa.theme import Theme, Style
from gappedbuf import re as gre
from kaa.command import Commands, command, norec, norerun
from kaa.keyboard import *

CThemes = {
    'basic': []
}

c_keys = {
    (ctrl, 't'): 'toc.showlist',
}


def build_tokenizer(stop=None, terminates=None):
    CTOKENS = namedtuple('ctokens',
                         ['macro', 'keywords', 'number', 'comment1', 'comment2', 'string1',
                          'string2', 'stop'])

    macro = SingleToken('c-macro', 'directive', [r'\#\s*\w+'])

    keywords = Keywords('c-keyword', 'keyword',
                        ['asm', 'auto', 'break', 'case', 'continue', 'const', 'char',
                         'default', 'do', 'double', 'else', 'endasm', 'entry', 'enum',
                         'extern', 'float', 'for', 'goto', 'if', 'int', 'long', 'register',
                         'return', 'short', 'signed', 'sizeof', 'static', 'struct',
                         'switch', 'typedef', 'union', 'unsigned', 'void', 'volatile',
                         'while'])

    number = SingleToken('c-numeric', 'number',
                         [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b'])
    comment1 = Span('c-comment1', 'comment', r'/\*', r'\*/', escape='\\')
    comment2 = Span('c-comment2', 'comment', r'//', r'$', escape='\\')
    string1 = Span('c-string1', 'string', '"', '"', escape='\\')
    string2 = Span('c-string2', 'string', "'", "'", escape='\\')

    tokens = CTOKENS(macro, keywords, number, comment1, comment2,
                     string1, string2, stop)

    return Tokenizer(tokens, terminates=terminates)


class CMode(defaultmode.DefaultMode):
    MODENAME = 'C'

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [c_keys])

    def init_themes(self):
        super().init_themes()
        self.themes.append(CThemes)

    def init_tokenizers(self):
        self.tokenizers = [build_tokenizer()]



import kaa


class Macro:
    recording = False
    commands = ()

    def start_record(self):
        self.recording = True
        self.commands = []

    def end_record(self):
        self.recording = False

    def toggle_record(self):
        self.recording = not self.recording
        if self.recording:
            self.commands = []

    def is_recording(self):
        return self.recording

    def get_commands(self):
        return self.commands

    def record(self, f, *args, **kwargs):
        if not hasattr(f, 'NOREC'):
            commandid = f.COMMAND_ID
            self.commands.append((commandid, args, kwargs))

    def record_string(self, s):
        if not self.commands or not isinstance(self.commands[-1], str):
            self.commands.append(s)
        else:
            self.commands[-1] += s

    def record_repeatcount(self, n):
        if not self.commands or not isinstance(self.commands[-1], (int, type(None))):
            self.commands.append(n)
        else:
            self.commands[-1] = n

    def run(self, wnd):
        if self.recording:
            return

        mode = wnd.document.mode
        for cmd in self.commands:
            if isinstance(cmd, str):
                wnd.document.mode.put_string(wnd, cmd)
                wnd.screen.selection.clear()
            elif isinstance(cmd, (int, type(None))):
                wnd.editmode.set_repeat(cmd)
            else:
                commandid, args, kwargs = cmd
                if callable(commandid):
                    commandid(wnd)
                else:
                    is_available, cmd = mode.get_command(commandid)
                    cmd(wnd, *args, **kwargs)



import sys
import os
KAA_VERSION = (0, 21, 0)


def version_info():
    return '''\
Kaa {major}.{minor}.{micro}
Copyright (c) 2013 Atsuo Ishimoto.
All Rights Reserved.

Python version: {python_version}
Platform: {platform}
$TERM: {term}
'''.format(
        major=KAA_VERSION[0], minor=KAA_VERSION[1],
        micro=KAA_VERSION[2],
        python_version=sys.version,
        platform=sys.platform,
        term=repr(os.environ.get('TERM', ''))[1:-1])



import itertools
import curses
import curses.panel
from kaa.cui.wnd import Window
import kaa
import kaa.log
from kaa import screen, cursor
from kaa.cui.color import ColorName


class TextEditorWindow(Window):

    """Text editor window"""

    OVERLAY_CURSORROW = 'cursor_row'  # Name of overlay style of theme

    show_lineno = True
    splitter = None
    document = None
    statusbar = None
    editmode = None
    visible = True
    highlight_cursor_row = False

    def _oninit(self):
        super()._oninit()

        self._cwnd.leaveok(0)
        self._cwnd.scrollok(0)

        self.screen = screen.Screen()
        self.screen.setsize(*self.getsize())
        self.cursor = cursor.Cursor(self)
        self.pending_str = ''
        self.charattrs = {}
        self.highlight_cursor_row = False
        self.line_overlays = {}

    def destroy(self):
        if self.document:
            self.document.del_window(self)
        self.screen.close()
        self.document = self.screen = self.cursor = None
        self.splitter = None
        self.statusbar = None

        super().destroy()

    def show_doc(self, doc):
        if doc is self.document:
            return

        if self.document:
            self.document.del_window(self)

        self._oninit()

        self.document = doc
        self.screen.set_document(doc)
        self.document.add_window(self)

        self.draw_screen()
        self.refresh()
        self.cursor.refresh(middle=True)

        kaa.app.set_idlejob()  # Reschedule idle procs

    def set_editmode(self, mode):
        self.editmode = mode
        self.editmode.activated(self)

    def dup(self):
        ret = self.__class__(parent=self.parent)
        return ret

    def set_splitter(self, splitter):
        self.splitter = splitter

    def set_statusbar(self, statusbar):
        self.statusbar = statusbar

    def set_highlight_cursor_row(self, f):
        ret = f != self.highlight_cursor_row
        self.highlight_cursor_row = f
        if ret:
            self.refresh()
        return ret

    def set_line_overlay(self, pos, overlay=None):
        if overlay is not None:
            if self.line_overlays.get(pos, '') != overlay:
                self.line_overlays[pos] = overlay
                self.screen.style_updated()
                return True
        else:
            if pos in self.line_overlays:
                del self.line_overlays[pos]
                self.screen.style_updated()
                return True

    def clear_line_overlay(self):
        if self.line_overlays:
            self.line_overlays = {}
            self.screen.style_updated()

    def bring_top(self):
        if curses.panel.top_panel() is not self._panel:
            super().bring_top()

            self.draw_screen(force=True)

    def activate(self):
        super().activate()
        kaa.app.set_focus(self)

    def set_cursor(self, cursor):
        self.cursor = cursor

    def set_visible(self, visible):
        self.visible = visible
        self.draw_screen(force=True)

    def _flush_pending_str(self):
        if self.editmode:
            return self.editmode.flush_pending_str(self)

    def _update_activeframe(self):
        frame = self.get_label('frame')
        if frame:
            frame.bring_top()
            frame.set_active_editor(self)

    def on_focus(self):
        super().on_focus()
        self._update_activeframe()

        if self.document:
            self.document.mode.on_focus(self)

    def _getcharattrs(self, row, rectangular, selfrom,
                      selto, colfrom, colto, line_overlay):
        # returns character attributes of each characters in row.

        ret = []
        if not row.positions:
            return ret
            
        posfrom = row.positions[0]
        styles = self.document.styles.getints(posfrom, row.positions[-1]+1)

        ncols = row.colfrom
        for pos, cols in zip(row.positions, row.cols):
            attr = 0
            sel = False
            if selfrom <= pos < selto:
                if rectangular:
                    if colfrom <= ncols < colto:
                        sel = True
                    else:
                        sel = False
                else:
                    sel = True

            if sel:
                attr = curses.A_REVERSE

            tokenid = self.charattrs.get(pos, None)
            if tokenid is None:
                tokenid = styles[pos-posfrom]
                
            style = self.document.mode.get_style(tokenid)
            color = None
            if line_overlay:
                color = style.cui_overlays.get(line_overlay, None)
            if color is None:
                color = style.cui_colorattr

            if style.underline:
                color += curses.A_UNDERLINE
            if style.bold:
                color += curses.A_BOLD

            ret.append((color + attr, style.rjust))

            ncols += cols
        return ret
        
    def draw_screen(self, force=False):
        try:
            self._draw_screen(force=force)
        except curses.error:
            kaa.log.debug('error on drawing: {}'.format(self), exc_info=True)

    def _draw_screen(self, force=False):
        # don't draw is frame is not visible.
        frame = self.get_label('frame')
        if frame:
            if kaa.app.get_activeframe() is not frame:
                return

        self.screen.apply_updates()
        self.screen.row_drawn()

        # save cursor position
        if kaa.app.focus:
            if self.document.mode.is_cursor_visible():
                cury, curx = kaa.app.focus._cwnd.getyx()

        h, w = self._cwnd.getmaxyx()

        rows = list(self.screen.get_visible_rows())
        cur_sel = self.screen.selection.get_selrange()

        lineno_width = 0
        lineno = self.document.buf.lineno.lineno(self.screen.pos)
        if self.document.mode.SHOW_LINENO:
            lineno_color = self.document.mode.theme.get_style(
                'lineno').cui_colorattr
            lineno_width = screen.calc_lineno_width(self.screen)

        _, cursorrow = self.screen.getrow(self.cursor.pos)

        tol = rows[0].tol
        eol = self.document.geteol(tol)
        endpos = self.document.endpos()

        rectangular = self.screen.selection.is_rectangular()
        selfrom = selto = colfrom = colto = -1

        if not rectangular:
            selrange = self.screen.selection.get_selrange()
            if selrange:
                selfrom, selto = selrange
        else:
            selrect = self.screen.selection.get_rect_range()
            if selrect:
                selfrom, selto, colfrom, colto = selrect

        theme = self.document.mode.theme
        defaultcolor = theme.get_style('default').cui_colorattr

        overlays = self.document.mode.get_line_overlays()
        overlays.update(self.line_overlays)

        for n, row in enumerate(rows):
            if n > h:
                break

            spos = 0

            if tol != row.tol:
                tol = row.tol
                eol = self.document.geteol(tol)

            line_overlay = None
            if row is cursorrow and (
                    self.highlight_cursor_row or
                    self.document.mode.HIGHLIGHT_CURSOR_ROW):

                line_overlay = self.OVERLAY_CURSORROW
            else:
                for pos in overlays.keys():
                    if (tol <= pos < eol) or (pos == tol == eol == endpos):
                        line_overlay = overlays[pos]
                        break

            # clear row
            self._cwnd.move(n, 0)
            self._cwnd.clrtoeol()
            if not line_overlay:
                erase_attr = defaultcolor
            else:
                style = theme.get_style('default')
                erase_attr = style.cui_overlays.get(line_overlay, defaultcolor)

            self._cwnd.chgat(n, 0, -1, erase_attr)

            if not self.visible:
                continue

            # draw line no
            if self.document.mode.SHOW_LINENO:
                self._cwnd.move(n, 0)

                if row.posfrom == row.tol:
                    self.add_str(str(lineno).rjust(lineno_width - 1),
                                 lineno_color)
                else:
                    self.add_str(' ' * (lineno_width - 1), lineno_color)

                self.add_str(' ', defaultcolor)

            # move cursor to top of row
            self._cwnd.move(n, row.wrapindent + lineno_width)
            rjust = False

            for (attr, attr_rjust), group in itertools.groupby(
                    self._getcharattrs(row, rectangular, selfrom, selto,
                                       colfrom, colto, line_overlay)):

                if not rjust and attr_rjust:
                    rjust = True

                    rest = sum(row.cols[spos:])
                    cy, cx = self._cwnd.getyx()
                    self._cwnd.move(cy, w - rest)

                slen = len(tuple(group))
                letters = ''.join(row.chars[spos:spos + slen]).rstrip('\n')
                self.add_str(letters, attr)
                spos += slen

            if row.chars == '\n':
                if selfrom <= row.posfrom < selto:
                    if not rectangular or (colfrom == 0):
                        self.add_str(' ', curses.A_REVERSE)


        if len(rows) < h:
            self._cwnd.move(len(rows), 0)
            self._cwnd.clrtobot()

        if kaa.app.focus:
            if self.document.mode.is_cursor_visible():
                kaa.app.focus._cwnd.move(cury, curx)

        return

    def on_document_updated(self, pos, inslen, dellen):
        self.screen.on_document_updated(pos, inslen, dellen)
        endpos = self.document.endpos()
        if self.cursor.pos > endpos:
            self.cursor.pos = endpos

    def style_updated(self, posfrom, posto):
        f, t = self.screen.get_visible_range()
        if posfrom <= t and f <= posto:
            self.screen.style_updated()

    CURSOR_TO_MIDDLE_ON_SCROLL = True

    def locate_cursor(self, pos, top=None, middle=None, bottom=None):

        if top is middle is bottom is None:
            middle = self.CURSOR_TO_MIDDLE_ON_SCROLL
            bottom = not self.CURSOR_TO_MIDDLE_ON_SCROLL

        self.screen.apply_updates()
        self.screen.locate(pos, top=top, middle=middle, bottom=bottom)

        idx, x = self.screen.getrowcol(pos)
        y = idx - self.screen.portfrom

        retpos = self.screen.get_pos_under(idx, x)

        screenx = x
        if self.document.mode.SHOW_LINENO:
            screenx = x + screen.calc_lineno_width(self.screen)

        if (y, screenx) != self._cwnd.getyx():
            h, w = self._cwnd.getmaxyx()
            if y < h and screenx < w and y >= 0 and screenx >= 0:
                self._cwnd.move(y, screenx)

        self.document.mode.on_cursor_located(self, retpos, y, x)
        if self.document.mode.HIGHLIGHT_CURSOR_ROW:
            self.screen.style_updated()

        return retpos, y, x

    def get_cursor_loc(self):
        y, x = self._cwnd.getyx()
        u, l = self._cwnd.getbegyx()
        return u + y, x + l

    def linedown(self):
        if self.screen.linedown():
            self._cwnd.scrollok(True)
            try:
                self._cwnd.scroll(1)
            finally:
                self._cwnd.scrollok(False)

            self.refresh()
            return True

    def lineup(self):
        if self.screen.lineup():
            self._cwnd.scrollok(True)
            try:
                self._cwnd.scroll(-1)
            finally:
                self._cwnd.scrollok(False)

            return True

    def pagedown(self):
        if self.screen.pagedown():
            return True

    def pageup(self):
        if self.screen.pageup():
            return True

    def on_setrect(self, l, t, r, b):
        if self.document:
            w = max(2, r - l)
            h = max(1, b - t)
            self.screen.setsize(w, h)
            self.draw_screen()
            self.cursor.setpos(self.cursor.pos)  # relocate cursor
            self.cursor.savecol()

            self.refresh()

    def update_window(self):
        # if this editor is a part of ChildFrame,
        # update only if the ChildFrame is active.
        frame = self.get_label('frame')
        if frame:
            if kaa.app.get_activeframe() is not frame:
                return

        self.screen.apply_updates()
        if self.screen.is_row_updated():
            if self.draw_screen():
                self.cursor.refresh()
                self.refresh()
                return True

    def on_killfocus(self):
        self._flush_pending_str()

    def on_idle(self):
        if not self.closed:
            if self._flush_pending_str():
                return True

            if self.document.mode.update_charattr(self):
                return True

            return self.update_status()

    def update_status(self):
        if self.statusbar:
            modified = False
            if self.document.undo:
                if self.document.undo.is_dirty():
                    modified = True

            linecount = self.document.buf.lineno.linecount()
            lineno = self.document.buf.lineno.lineno(self.cursor.pos)
            col = self.screen.get_cursorcol(self.cursor.pos)
            updated = self.statusbar.set_info(
                filename=self.document.get_title(),
                modified=modified,
                lineno=lineno,
                col=col + 1,
                linecount=linecount,
                editmode=self.editmode.MODENAME,
            )
            return updated



import os
import collections
import fnmatch
import kaa
from kaa import document, encodingdef, utils, consts
from kaa.ui.dialog import dialogmode
from kaa.ui.selectlist import selectlist
from kaa.ui.itemlist import itemlistmode
from kaa.ui.inputline import inputlinemode
from kaa.ui.msgbox import msgboxmode

from kaa.theme import Theme, Style

from kaa.command import Commands, command, norec, norerun
from kaa.keyboard import *
from kaa.commands import editorcommand
import kaa.filetype.default.keybind


FileOpenDlgThemes = {
    'basic': [],
}


class DirFileListMode(selectlist.SelectItemList):
    dirname = None
    filename = None
    dirs = files = ()

    def init_themes(self):
        super().init_themes()
        self.themes.append(FileOpenDlgThemes)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)
        wnd.set_label('filelist', wnd)

    def set_dir(self, dirname):
        self.caption = self.dirname = os.path.abspath(
            os.path.expanduser(dirname))
        self.read_dir()

    def read_dir(self):
        try:
            self.dirs, self.files = kaa.app.storage.listdir(self.dirname)
        except OSError as e:
            kaa.app.messagebar.set_message(str(e))
            self.dirs = []
            self.files = []

        # add '..' if dir is not root
        if self.dirname != '/':
            if os.path.exists(os.path.join(self.dirname, '..')):
                self.dirs.insert(0, '..')

        self.dirs = [name + '/' for name in self.dirs]

    def set_filename(self, filename):
        if filename.strip():
            upper = filename.strip().upper().split()

            def f(s):
                s = s.upper()
                for u in upper:
                    if u not in s:
                        return False
                return True
            self.filterfunc = f
        else:
            self.filterfunc = None

    def show_files(self, wnd):
        self.cursel = None
        dirs = [selectlist.SelectItem(
            'selectitem', 'selectitem-active', name, name) for name in self.dirs]
        files = [selectlist.SelectItem(
            'selectitem2', 'selectitem-active', name, name) for name in self.files]
        items = dirs + files
        if self.filterfunc:
            items = [item for item in items if self.filterfunc(item.text)]
        self.update_doc(items)

FileNameDlgThemes = {
    'basic': [],
}


fileopendlg_keys = {
    tab: 'fileopendlg.next',
    (shift, tab): 'fileopendlg.prev',
    '\r': 'fileopendlg.openfile',
    '\n': 'fileopendlg.openfile',
}


class FileOpenDlgCommands(Commands):

    def _update_filefield(self, wnd, filename):
        if filename and wnd.document.mode.get_filename() != filename:
            wnd.document.mode.set_filename(wnd, filename)

    def _check_needcomplete(self, wnd):
        cur = wnd.document.mode.get_filename()
        cur = os.path.expanduser(cur)

        if os.path.isabs(cur):
            if os.path.isdir(cur):
                cur += os.sep
            self.show_filename(wnd, cur)
            return True

        d, f = os.path.split(cur)
        if d and f:
            filename = self._build_filename(wnd)
            self.show_filename(wnd, filename)
        else:
            d, f = os.path.split(d)
            if d and f:
                filename = self._build_filename(wnd)
                self.show_filename(wnd, filename)

    @command('fileopendlg.next')
    @norec
    @norerun
    def next(self, wnd):
        if self._check_needcomplete(wnd):
            return

        filelist = wnd.get_label('filelist')
        cursel = filelist.document.mode.sel_next(filelist)
        if cursel:
            self._update_filefield(wnd, cursel.value)

    @command('fileopendlg.prev')
    @norec
    @norerun
    def prev(self, wnd):
        if self._check_needcomplete(wnd):
            return

        filelist = wnd.get_label('filelist')
        cursel = filelist.document.mode.sel_prev(filelist)
        if cursel:
            self._update_filefield(wnd, cursel.value)

    def _build_filename(self, wnd):
        filename = wnd.document.mode.get_filename()
        filename = os.path.expanduser(filename)

        if not os.path.isabs(filename):
            filelist = wnd.get_label('filelist')
            filename = os.path.join(filelist.document.mode.dirname, filename)

            if os.path.isdir(filename):
                filename += os.path.sep

        return filename

    def show_filename(self, wnd, filename):
        # split filename into existing sub-directories and rest of filename.
        dir, rest = utils.split_existing_dirs(filename)

        # set existing directory as current dir.
        filelist = wnd.get_label('filelist')
        filelist.document.mode.set_dir(dir)

        # set rest of filename to filename filter.
        filelist.document.mode.set_filename(rest)

        filelist.document.mode.show_files(wnd)
        filelist.get_label('popup').on_console_resized()

        # set rest of filename filename field.
        wnd.document.mode.set_filename(wnd, rest)

    @command('fileopendlg.openfile')
    @norec
    @norerun
    def openfile(self, wnd):
        filename = self._build_filename(wnd)
        if os.path.isfile(filename):
            enc = wnd.document.mode.encoding
            if enc == 'japanese':
                enc = kaa.app.storage.guess_japanese_encoding(filename)
                if not enc:
                    kaa.app.messagebar.set_message(
                        'Cannot detect text encoding.')
                    return

            wnd.document.mode.callback(filename, enc,
                                       wnd.document.mode.newline)
            popup = wnd.get_label('popup')
            popup.destroy()
            return

        if os.path.isdir(filename):
            if not filename.endswith(os.path.sep):
                filename += os.path.sep
        else:
            kaa.app.messagebar.set_message('{} is not found.'.format(filename))

        self.show_filename(wnd, filename)


class OpenFilenameDlgMode(dialogmode.DialogMode):
    MAX_INPUT_HEIGHT = 4
    autoshrink = True

    @classmethod
    def build(cls, filename, newline, encoding, callback):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)

        mode.newline = newline if newline else kaa.app.config.DEFAULT_NEWLINE
        mode.encoding = encoding if encoding else kaa.app.config.DEFAULT_ENCODING
        mode.callback = callback

        with dialogmode.FormBuilder(doc) as f:
            f.append_text('caption', 'Filename:')
            f.append_text('default', ' ')
            f.append_text('default', filename, mark_pair='filename')
            f.append_text('default', ' ')

            f.append_text('checkbox', '[&Encoding:{}]'.format(mode.encoding),
                          mark_pair='enc',
                          shortcut_style='checkbox.shortcut',
                          on_shortcut=lambda wnd:
                          wnd.document.mode.select_encoding(wnd))

            f.append_text('checkbox', '[&Newline:{}]'.format(mode.newline),
                          mark_pair='newline',
                          shortcut_style='checkbox.shortcut',
                          on_shortcut=lambda wnd:
                          wnd.document.mode.select_newline(wnd))

            f.append_text('checkbox', '[&Create dir]',
                          shortcut_style='checkbox.shortcut',
                          on_shortcut=lambda wnd:
                          wnd.document.mode.create_dir(wnd))

        return doc

    def close(self):
        super().close()

    def init_themes(self):
        super().init_themes()
        self.themes.append(FileNameDlgThemes)

    def init_keybind(self):
        self.keybind.add_keybind(
            kaa.filetype.default.keybind.edit_command_keys)
        self.keybind.add_keybind(kaa.filetype.default.keybind.cursor_keys)
        self.keybind.add_keybind(kaa.filetype.default.keybind.emacs_keys)
        self.keybind.add_keybind(fileopendlg_keys)

    def init_commands(self):
        super().init_commands()

        self.fileopendlg_commands = FileOpenDlgCommands()
        self.register_command(self.fileopendlg_commands)

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

    def create_cursor(self, wnd):
        return dialogmode.DialogCursor(wnd,
                                       [dialogmode.MarkRange('filename')])

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        wnd.CURSOR_TO_MIDDLE_ON_SCROLL = False
        wnd.set_cursor(self.create_cursor(wnd))
        wnd.cursor.setpos(self.document.marks['filename'][1])

        wnd.set_label('filename_field', self)
        kaa.app.messagebar.set_message("Hit tab/shift+tab to complete.")

    def on_edited(self, wnd):
        super().on_edited(wnd)

        filename = wnd.document.mode.get_filename()
        if os.sep not in filename:
            filelist = wnd.get_label('filelist')
            filelist.document.mode.set_filename(filename)
            filelist.document.mode.show_files(wnd)
            filelist.get_label('popup').on_console_resized()

    def calc_position(self, wnd):
        w, h = wnd.getsize()
        height = self.calc_height(wnd)
        height = min(height, self.MAX_INPUT_HEIGHT)
        top = wnd.mainframe.height - height - wnd.mainframe.MESSAGEBAR_HEIGHT
        return 0, top, wnd.mainframe.width, top + height

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("")
        self.callback(None, None, None)

    def get_filename(self):
        f, t = self.document.marks['filename']
        return self.document.gettext(f, t)

    def set_filename(self, wnd, s):
        f, t = self.document.marks['filename']
        self.document.replace(f, t, s)
        wnd.screen.selection.set_range(f, f + len(s))
        wnd.cursor.setpos(f + len(s))

    def _get_encnames(self):
        return sorted(encodingdef.encodings + ['japanese'],
                      key=lambda k: k.upper())

    def select_encoding(self, wnd):
        encnames = self._get_encnames()

        def callback(n):
            if n is None:
                return

            enc = encnames[n]
            if enc != self.encoding:
                self.encoding = enc
                f, t = self.document.marks['enc']
                # [Encoding:{mode}]
                # 01234567890    10
                self.document.replace(f + 10, t - 1, self.encoding)

        doc = itemlistmode.ItemListMode.build(
            'Select character encoding:',
            encnames,
            encnames.index(self.encoding),
            callback)

        kaa.app.show_dialog(doc)

    def select_newline(self, wnd):
        def callback(n):
            if n is None:
                return

            nl = consts.NEWLINES[n]
            if nl != self.newline:
                self.newline = nl
                f, t = self.document.marks['newline']
                # [Newline:{mode}]
                # 0123456789    10
                self.document.replace(f + 9, t - 1, self.newline)

        doc = itemlistmode.ItemListMode.build(
            'Select newline mode:',
            consts.NEWLINES,
            consts.NEWLINES.index(self.newline),
            callback)

        kaa.app.show_dialog(doc)

    def create_dir(self, wnd):
        filelist = wnd.get_label('filelist')
        dirname = filelist.document.mode.dirname

        def callback(w, path):
            path = os.path.join(dirname, path)
            try:
                os.makedirs(path)
            except Exception as e:
                kaa.app.messagebar.set_message(str(e))
            else:
                filelist = wnd.get_label('filelist')
                filelist.document.mode.read_dir()
                filelist.document.mode.show_files(wnd)

        doc = inputlinemode.InputlineMode.build('directory name:', callback)
        kaa.app.show_dialog(doc)


def show_fileopen(filename, callback):
    if not filename:
        filename = kaa.app.last_dir

    filename = os.path.abspath(
        os.path.expanduser(filename))

    if os.path.isdir(filename) and not filename.endswith(os.path.sep):
        filename += os.path.sep

    doc = OpenFilenameDlgMode.build('', None, None, callback)
    dlg = kaa.app.show_dialog(doc)

    filelist = DirFileListMode.build()
    dlg.add_doc('dlg_filelist', 0, filelist)

    doc.mode.fileopendlg_commands.show_filename(
        dlg.get_label('editor'), filename)
    dlg.on_console_resized()

    return doc


class FileSaveAsDlgCommands(FileOpenDlgCommands):

    @command('fileopendlg.openfile')
    @norec
    @norerun
    def openfile(self, wnd):
        filename = self._build_filename(wnd)

        # change current directory if filename is existing directory.
        if not os.path.isdir(filename):

            # Is filename contain valid filename?
            dir, rest = utils.split_existing_dirs(filename)
            if rest and os.sep not in rest:

                # query overwrite if file already exists.
                if os.path.exists(filename):
                    def choice(c):
                        if c in 'yY':
                            wnd.document.mode.callback(
                                filename, wnd.document.mode.encoding,
                                wnd.document.mode.newline)
                            wnd.get_label('popup').destroy()

                    msgboxmode.MsgBoxMode.show_msgbox(
                        'File `{}` already exists. Overwrite?: '.format(
                            filename),
                        ['&Yes', '&No'], choice)
                else:
                    wnd.document.mode.callback(
                        filename, wnd.document.mode.encoding,
                        wnd.document.mode.newline)
                    wnd.get_label('popup').destroy()
                return

        return super().openfile(wnd)


class SaveAsFilenameDlgMode(OpenFilenameDlgMode):

    def init_commands(self):
        super().init_commands()

        self.fileopendlg_commands = FileSaveAsDlgCommands()
        self.register_command(self.fileopendlg_commands)

    def _get_encnames(self):
        return sorted(encodingdef.encodings,
                      key=lambda k: k.upper())


def show_filesaveas(filename, encoding, newline, callback):
    if not filename:
        filename = kaa.app.last_dir

    filename = os.path.abspath(
        os.path.expanduser(filename))
    if os.path.isdir(filename) and not filename.endswith(os.path.sep):
        filename += os.path.sep

    doc = SaveAsFilenameDlgMode.build('', encoding, newline, callback)
    dlg = kaa.app.show_dialog(doc)

    filelist = DirFileListMode.build()
    dlg.add_doc('dlg_filelist', 0, filelist)

    doc.mode.fileopendlg_commands.show_filename(
        dlg.get_label('editor'), filename)

    return doc


class SelectDirDlgMode(OpenFilenameDlgMode):

    @classmethod
    def build(cls, filename, callback):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)

        mode.callback = callback

        with dialogmode.FormBuilder(doc) as f:
            f.append_text('caption', 'Directory name:')
            f.append_text('default', ' ')
            f.append_text('default', filename, mark_pair='filename')
            f.append_text('default', ' ')

            f.append_text('checkbox', '[&Select this dir]',
                          shortcut_style='checkbox.shortcut',
                          on_shortcut=lambda wnd:
                          wnd.document.mode.on_select_dir(wnd))

        return doc

    def on_esc_pressed(self, wnd, event):
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("")
        self.callback(None)

    def on_select_dir(self, wnd):
        filename = self.fileopendlg_commands._build_filename(wnd)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("")
        self.callback(filename)


class DirListMode(DirFileListMode):

    def show_files(self, wnd):
        self.cursel = None
        dirs = [selectlist.SelectItem(
            'selectitem', 'selectitem-active', name, name) for name in self.dirs]
        if self.filterfunc:
            dirs = [item for item in dirs if self.filterfunc(item.text)]
        self.update_doc(dirs)


def show_selectdir(curdir, callback):
    doc = SelectDirDlgMode.build('', callback)
    dlg = kaa.app.show_dialog(doc)

    filelist = DirListMode.build()
    dlg.add_doc('dlg_filelist', 0, filelist)

    doc.mode.fileopendlg_commands.show_filename(
        dlg.get_label('editor'), curdir)

    return doc



import itertools
import re
import math
from unicodedata import east_asian_width

import kaa
from kaa import document
from kaa.document import is_combine


def calc_lineno_width(screen):
    if not screen.rows:
        return 0

    lineno = screen.document.buf.lineno.lineno(screen.rows[-1].posfrom)
    digits = int(math.log10(lineno) + 1)
    digits += 2     # one space left and right of line number.
    return digits


class Row:
    height = 1
    bgcolor = None

    def __init__(self, posfrom, tol, colfrom, wrapindent, chars, cols,
                 positions, intervals):
        self.posfrom = posfrom
        self.posto = positions[-1] + 1 if positions else posfrom
        self.colfrom = colfrom
        self.tol = tol
        self.wrapindent = wrapindent
        self.chars = chars
        self.cols = cols
        self.positions = positions
        self.intervals = intervals

    def get_char(self, pos):
        ret = []
        for p, c in zip(self.positions, self.chars):
            if p == pos:
                ret.append(c)
        return ''.join(ret)


def translate_chars(posfrom, chars, tab_width, ambiguous_width):
    """Build character informations from chars.

    Returns tuple of four lists. The first item is a list of characters to be
    displayed. Second item is a list of columns of each characters. Third item
    is a list of position in document of each characters. Last item is a list of
    interval from first column top of character of each character.
    """

    curcol = 0
    dispchrs = ''
    dispcols = []
    positions = []
    intervals = []
    double_width = set('WF')
    if ambiguous_width == 2:
        double_width.add('A')

    pos = posfrom
# return chars, [1]*len(chars), [(posfrom+i) for i in range(len(chars))],
# [0]*len(chars)

    for c in chars:
        if c == '\t':
            dispstr = ' ' * (tab_width - (curcol % tab_width))
        elif (((c != '\n') and (c < '\x20'))  # control chars
              or (c == '\x7f')                     # backspace
              or ('\ud800' <= c <= '\udfff')):     # surrogate pair
            dispstr = repr(c)[1:-1]
        else:
            dispstr = c

        dispchrs += dispstr
        for i, d in enumerate(dispstr):
            positions.append(pos)
            intervals.append(i)
            # first char of line never be combined
            if curcol and is_combine(c):
                cols = 0
            else:
                w = east_asian_width(d)
                cols = 2 if w in double_width else 1
            dispcols.append(cols)
            curcol += cols
        pos += 1

# return chars, [1]*len(chars), [(posfrom+i) for i in range(len(chars))],
# [0]*len(chars)
    return dispchrs, dispcols, positions, intervals


MIN_WRAPCOLS = 10


def col_splitter(maxcol, tol, dispchrs, dispcols, positions,
                 intervals, styles, stylemap, nowrap=False, nowrapindent=False):
    """Split string by column"""

    if nowrap:
        return [Row(tol, tol, 0, 0, dispchrs, dispcols, positions, intervals)]

    assert maxcol >= 2
    rowfrom = rowto = 0
    sumcols = 0
    wrappos = None
    sumcols_at_wrappos = 0

    posfrom = tol
    colfrom = 0
    wrapindent = 0

    ret = []

    for col in dispcols:

        if not col:
            rowto += 1
            continue

        # check if nowrap
        if rowfrom < rowto:
            wrappable = True

            if rowto < len(positions):

                curpos = positions[rowto - 1]
                nextpos = positions[rowto]

                tokenid = styles[curpos - tol]
                nexttokenid = styles[nextpos - tol]

                style = stylemap.get(tokenid, None)
                if style and style.nowrap:
                    nextstyle = stylemap.get(nexttokenid)
                    if nextstyle and nextstyle.nowrap:
                        wrappable = False

            if wrappable:
                wrappos = rowto
                sumcols_at_wrappos = sumcols

        if ((sumcols + col) >= (maxcol - wrapindent)) and (rowfrom != rowto):
            # do not split at newlines
            if (rowto >= len(dispchrs)) or (dispchrs[rowto] != '\n'):

                if wrappos is None:
                    # No wrappable position found. So, fill as many chars as we
                    # can.
                    wrappos = rowto
                    sumcols_at_wrappos = sumcols

                assert rowfrom != wrappos
                row = Row(
                    posfrom, tol, colfrom, wrapindent, dispchrs[
                        rowfrom:wrappos],
                    dispcols[rowfrom:wrappos], positions[
                        rowfrom:wrappos],
                    intervals[rowfrom:wrappos])

                ret.append(row)
                colfrom += sum(dispcols[rowfrom:wrappos])

                rowfrom = wrappos
                sumcols = sumcols - sumcols_at_wrappos

                wrappos = None
                sumcols_at_wrappos = None

                posfrom = row.posto

                # set wrap-indent width after second row
                if not nowrapindent and len(ret) == 1:
                    wrapindent = re.match(r' *', dispchrs).end()
                    if wrapindent + MIN_WRAPCOLS > maxcol:
                        wrapindent = max(0, maxcol - MIN_WRAPCOLS)

        sumcols += col
        rowto += 1

    if rowfrom != len(dispcols) or not dispchrs:
        row = Row(
            posfrom, tol, colfrom, wrapindent, dispchrs[
                rowfrom:], dispcols[rowfrom:],
            positions[rowfrom:], intervals[rowfrom:])
        ret.append(row)

    return ret


class Selection:

    def __init__(self, screen):
        self._marks = document.Marks()
        self.screen = screen
        self._rectangular = False

    def _get_cursor_start(self):
        return self._marks.get('start', None)

    def _set_cursor_start(self, pos):
        self._marks['start'] = pos

    def _is_cursor_started(self):
        """Return True if range was selected"""

        return self._get_cursor_start() is not None

    def get_end(self):
        return self._marks.get('end', None)

    def _set_end(self, pos):
        self._marks['end'] = pos

    def _set_mark(self, pos):
        self._marks['mark'] = pos

    def _get_mark(self):
        return self._marks.get('mark', None)

    def has_mark(self):
        return self._get_mark() is not None

    def is_selected(self):
        if self._is_cursor_started() or self.has_mark():
            sel = self.get_selrange()
            if sel:
                f, t, = sel
                return f != t

    def is_rectangular(self):
        return self._rectangular

    def set_mark(self, pos):
        cur = self.get_selrange()
        self._rectangular = False

        self._set_mark(pos)
        if pos is not None:
            self._set_cursor_start(None)
        if cur != self.get_selrange():
            self.screen.style_updated()
            return True

    def set_rectangle_mark(self, pos):
        ret = self.set_mark(pos)
        self._rectangular = True

    def begin_cursor(self, pos):
        """Start cursor range selection if it was not started"""

        if self.has_mark():
            return

        if self._is_cursor_started():
            return

        self._rectangular = False
        self._set_cursor_start(pos)
        self._set_end(pos)

    def set_to(self, pos):
        """Update where selection to."""

        if not self.has_mark() and not self._is_cursor_started():
            return
        changed = self.get_end() != pos
        self._set_end(pos)

        if changed:
            self.screen.style_updated()

    def end_cursor(self):
        """Clear selection"""

        changed = self._is_cursor_started()

        self._set_cursor_start(None)
        self._set_end(None)

        if changed:
            self.screen.style_updated()

    def clear(self):
        """Clear selection"""

        changed = self._is_cursor_started() is not None
        if not changed:
            changed = self._get_mark() is not None

        self._set_cursor_start(None)
        self._set_end(None)
        self._set_mark(None)

        if changed:
            self.screen.style_updated()

    def get_selrange(self):
        start = self._get_cursor_start()
        if start is None:
            start = self._get_mark()

        end = self.get_end()
        if start is None or end is None or (start == end):
            return None

        return tuple(sorted((start, end)))

    def set_range(self, f, t):
        self._rectangular = False
        start = self._get_cursor_start()
        end = self.get_end()
        if (start, end) != (f, t):
            self._set_cursor_start(f)
            self._set_end(t)
            self._set_mark(None)
            self.screen.style_updated()

    def _calc_col(self, pos):
        tol = self.screen.document.gettol(pos)
        eol, s = self.screen.document.getline(tol)
        (dispchrs, dispcols, positions, intervals) = translate_chars(
            tol, s, self.screen.document.mode.tab_width,
            kaa.app.config.AMBIGUOUS_WIDTH)

        if pos < eol:
            n = positions.index(pos)
        else:
            n = len(dispcols)
        return tol, eol, sum(dispcols[0:n])

    def get_rect_range(self):
        sel = self.get_selrange()
        if not sel:
            return
        posfrom, _, col1 = self._calc_col(sel[0])
        _, posto, col2 = self._calc_col(sel[1])
        colfrom, colto = (col1, col2) if (col1 < col2) else (col2, col1)

        return posfrom, posto, colfrom, colto

    def get_col_string(self, tol, colfrom, colto):
        eol, s = self.screen.document.getline(tol)
        (dispchrs, dispcols, positions, intervals) = translate_chars(
            tol, s, self.screen.document.mode.tab_width,
            kaa.app.config.AMBIGUOUS_WIDTH)

        col = 0
        for top, c in enumerate(dispcols):
            if colfrom <= col:
                top -= intervals[top]
                break
            col += c
        else:
            return

        for end, c in enumerate(dispcols[top:]):
            if colto <= col:
                end = top + end
                end -= intervals[end]
                break
            col += c
        else:
            end = len(dispcols)

        toppos = positions[top] if top < len(dispcols) else eol
        endpos = positions[end] if end < len(dispcols) else eol
        return toppos, endpos, s[toppos - tol:endpos - tol]

    def on_document_updated(self, pos, inslen, dellen):
        self._marks.updated(pos, inslen, dellen)


class Screen:

    """
    Attributes:
        rows --  list of Row objects.

        portfrom, portto -- Index of portion of rows to be displayed
            on the screen.
    """

    def __init__(self):
        self._oninit()
        self.width = 2      # width of screen
        self.height = 1     # height of screen

    def _oninit(self):
        self.document = None
        self.nowrap = False
        self.build_entire_rows = False

        # Rows displayed on screen
        # A row at self.rows[0] should be start from top
        # of document or from a top of physical line.
        # A row at self.rows[-1] should be finished at end
        # of document or an end of physical line.
        # An entire physical line will be stored in self.rows,
        # never be a part of physical line.
        self.rows = []
        self.portfrom = 0   # self.rows[self.portfrom:self.portto]
                            # returns rows displayed on the screen.
        self.portto = 0
        self.pos = 0        # Position of top-left corner
        self.updated_pos = None
        self._style_updated = False
        self._need_redraw = False
        
        self.selection = Selection(self)

    def set_document(self, doc):
        self._oninit()

        self.document = doc
        self.nowrap = doc.mode.SCREEN_NOWRAP
        self.build_entire_rows = doc.mode.SCREEN_BUILD_ENTIRE_ROWS

    def close(self):
        self.document = self.selection = None

    def get_visible_rows(self):
        return self.rows[self.portfrom:self.portto]

    def get_visible_range(self):
        if not self.rows:
            return
        rows = self.get_visible_rows()
        if not rows:
            return -1, -1
        return (rows[0].posfrom, rows[-1].posto)

    def get_total_height(self):
        self.apply_updates()

        if self.build_entire_rows:
            if self.rows:
                assert self.rows[0].posfrom == 0

        return sum(r.height for r in self.rows)

    def setsize(self, width, height):
        if self.width != width or self.height != height:
            self.width = max(2, width)
            self.height = max(1, height)
            if self.document:
                self.locate(self.pos, top=True, refresh=True)

    def _buildrow(self, pos, s, styles):
        dispchrs, dispcols, positions, intervals = translate_chars(
            pos, s, self.document.mode.tab_width,
            kaa.app.config.AMBIGUOUS_WIDTH)

        if self.document.mode.SHOW_LINENO:
            linenowidth = calc_lineno_width(self)
        else:
            linenowidth = 0

        nowrapindent = self.document.mode.NO_WRAPINDENT
        width = max(2, self.width - linenowidth)
        return col_splitter(width, pos, dispchrs, dispcols,
                            positions, intervals, styles, self.document.mode.stylemap, nowrap=self.nowrap, nowrapindent=nowrapindent)

    def _set_rowport(self):

        # Update bottom of visible rows
        height = 0
        for i in range(self.portfrom, len(self.rows)):
            if height >= self.height:
                self.portto = i
                break
            height += self.rows[i].height
        else:
            self.portto = len(self.rows)

        # Remove unnecessary rows
        if not self.build_entire_rows:
            porttop = self.rows[self.portfrom].tol
            for i, row in enumerate(self.rows):
                if row.tol == porttop:
                    del self.rows[0:i]
                    self.portfrom -= i
                    self.portto -= i
                    break

            portend = self.rows[self.portto - 1].tol
            for i, row in enumerate(self.rows[self.portto:]):
                if row.tol != portend:
                    del self.rows[self.portto + i:]
                    break

    def on_document_updated(self, pos, inslen, dellen):
        self.selection.on_document_updated(pos, inslen, dellen)

        if not self.rows:
            return

        # calculate new top of screen position after update
        if pos > self.rows[-1].posto:
            # Nothing changed on screen
            return

        if self.updated_pos is None:
            self.updated_pos = self.rows[self.portfrom].posfrom

        if inslen >= dellen:
            if pos < self.updated_pos:
                self.updated_pos += inslen - dellen
        elif inslen < dellen:
            deleted = dellen - inslen
            if pos < self.updated_pos:
                if pos + deleted > self.updated_pos:
                    self.updated_pos = pos
                else:
                    tol = self.document.gettol(self.updated_pos - deleted)
                    if tol != self.document.gettol(pos):
                        # Position of top of screen is not changed
                        # if deleted on same physical line.
                        self.updated_pos -= deleted

        self.updated_pos = min(self.updated_pos, self.document.endpos())

    def style_updated(self):
        self._style_updated = True
#        if self.updated_pos is None:
#            if len(self.rows) > self.portfrom:
#                self.updated_pos = self.rows[self.portfrom].posfrom
            
    def apply_updates(self):
        if not self.rows:
            self.locate(0, top=True)
            return True

        if self.updated_pos is not None:
            ret = self.locate(self.updated_pos,
                              top=True, refresh=True) or ret
        self.updated_pos = None

        if self._style_updated:
            self._need_redraw = True
            self._style_updated = False

    def is_row_updated(self):
        return self._need_redraw

    def row_drawn(self):
        self._need_redraw = False

    def is_lastrow(self, row):
        if row.posto == self.document.endpos():
            if not row.chars:
                return True
            if not row.chars[-1].endswith('\n'):
                return True

    def is_visible(self, pos):
        idx, col = self.getrowcol(pos)
        if idx == -1:
            return False
        return self.portfrom <= idx < self.portto

    def get_pos_under(self, rowidx, col):
        """Get pos at col of row when cursor downed from above row"""

        row = self.rows[rowidx]
        pos = self._getpos_fromrowcol(row, col)

        # Return next pos if the pos start at a row above
        if row.cols:
            if (row.positions[0] == pos) and (row.intervals[0] != 0):
                pos = self.document.get_nextpos(pos)

        return pos

    def get_pos_above(self, rowidx, col):
        """Get pos at col of row when cursor upped from below row"""

        row = self.rows[rowidx]
        pos = self._getpos_fromrowcol(row, col)

        # Return next pos if the pos start at a row above
        if row.cols:
            if (row.positions[0] == pos) and (row.intervals[0] != 0):
                nextpos = self.document.get_nextpos(pos)
                if nextpos < row.positions[-1]:
                    return nextpos

        return pos

    def _getpos_fromrowcol(self, row, col):
        """Returns position of specified column"""

        if not row.cols:
            return row.posfrom

        p = 0
        col = col - row.wrapindent
        for pos, charcols, c in zip(row.positions, row.cols, row.chars):
            if (p + charcols > col) or (c == '\n'):
                return pos
            p += charcols
        else:
            if self.is_lastrow(row):
                return row.posto
            else:
                ret = row.positions[-1]
                return ret

    def getrow(self, pos):
        assert pos is not None
        if self.rows and self.rows[0].posfrom <= pos:
            for i, row in enumerate(self.rows):
                if row.posfrom <= pos < row.posto:
                    return i, row
            if self.rows and self.is_lastrow(self.rows[-1]):
                return len(self.rows) - 1, self.rows[-1]

        return -1, None

    def getrowcol(self, pos):
        idx, row = self.getrow(pos)
        if idx == -1:
            return -1, -1
        else:
            try:
                col = row.positions.index(pos)
            except ValueError:
                col = len(row.positions)
            return idx, sum(c for c in row.cols[:col]) + row.wrapindent

    def get_cursorcol(self, pos):
        idx, row = self.getrow(pos)
        if idx == -1:
            return 0

        if row.cols:
            if pos == row.posto:
                cols = sum(row.cols)
            else:
                cols = sum(row.cols[:row.positions.index(pos)])
        else:
            cols = 0

        for n in range(idx - 1, -1, -1):
            nrow = self.rows[n]
            if nrow.tol != row.tol:
                return cols
            else:
                cols += sum(nrow.cols)

        return cols

    def get_pos_at_cols(self, tol, cols):
        posidx, row = self.getrow(tol)
        if not row:
            return tol

        p = 0
        for n in range(posidx, len(self.rows)):
            row = self.rows[n]
            if row.tol != tol:
                break

            for pos, charcols, c in zip(row.positions, row.cols, row.chars):
                if p + charcols > cols:
                    return pos
                p += charcols

        return self.document.find_newline(tol)

    def _fillscreen(self):
        while True:
            bottomrow = self.rows[-1]
            if self.is_lastrow(bottomrow):
                break

            height = sum(row.height for row in self.rows[self.portfrom:])
            if not self.build_entire_rows and (height >= self.height):
                break

            eol, s = self.document.getline(bottomrow.posto)
            styles = self.document.getstyles(bottomrow.posto, eol)
            self.rows.extend(self._buildrow(bottomrow.posto, s, styles))

        self._set_rowport()

    def locate(self, pos, top=False, middle=False, bottom=False,
               align_always=False, refresh=False):

        assert top or middle or bottom
        if self.updated_pos is not None:
            refresh = True
            self.updated_pos = None

        posidx = -1
        if not refresh:
            posidx, posrow = self.getrow(pos)

        if not refresh and (posidx != -1):
            if not align_always and (self.portfrom <= posidx < self.portto):
                return False

        elif self.build_entire_rows:
            eol, s = self.document.getline(0)
            styles = self.document.getstyles(0, eol)
            self.rows = self._buildrow(0, s, styles)
            self._fillscreen()
            posidx, posrow = self.getrow(pos)

        else:
            # build specified row
            tol = self.document.gettol(pos)
            eol, s = self.document.getline(tol)
            styles = self.document.getstyles(tol, eol)
            self.rows = self._buildrow(tol, s, styles)
            posidx, posrow = self.getrow(pos)

        self._need_redraw = True
        self.vert_align(posidx, top, middle, bottom)

        return True

    def vert_align(self, rowidx, top=False, middle=False, bottom=False):
        assert top or middle or bottom
        # move the row to middle or bottom

        old_range = self.get_visible_range()
        targetrow = self.rows[rowidx]

        if top:
            rowbottom = 0
        elif middle:
            rowbottom = self.height // 2
        else:
            # bottom
            rowbottom = max(0, self.height - targetrow.height)

        if rowbottom > 0:
            # build rows above target row
            height = sum(row.height for row in self.rows[:rowidx + 1])

            while height <= rowbottom:
                curtop = self.rows[0].posfrom
                if curtop == 0:
                    # top of buffer
                    break

                # build previous line
                top = self.document.gettol(curtop - 1)
                eol, s = self.document.getline(top)
                styles = self.document.getstyles(top, eol)
                rows = self._buildrow(top, s, styles)
                height += sum(row.height for row in rows)
                rowidx += len(rows)

                self.rows[0:0] = rows

            # move row
            height = 0
            for n in range(rowidx, -1, -1):
                if height > rowbottom:
                    break
                row = self.rows[n]
                height += row.height
                rowidx = n

        self.portfrom = rowidx
        self.pos = self.rows[rowidx].posfrom

        self._fillscreen()

        ret = old_range != self.get_visible_range()
        if ret:
            self._need_redraw = True
        return ret

    def linedown(self):
        if self.portfrom < len(self.rows) - 1:
            self.portfrom += 1
            self.pos = self.rows[self.portfrom].posfrom
            self._fillscreen()
            self._need_redraw = True
            return True
        else:
            currow = self.rows[self.portfrom]
            if not self.is_lastrow(currow):
                self.locate(currow.posto, top=True)
                self._need_redraw = True
                return True
        return False

    def lineup(self):
        if self.portfrom > 0:
            self.portfrom -= 1
            self.pos = self.rows[self.portfrom].posfrom
            self._fillscreen()
            self._need_redraw = True
            return True

        elif self.pos > 0:
            tol = self.document.gettol(self.pos - 1)
            eol, s = self.document.getline(tol)
            styles = self.document.getstyles(tol, eol)
            rows = self._buildrow(tol, s, styles)
            self.rows[0:0] = rows
            self.pos = rows[-1].posfrom
            self.portfrom = len(rows) - 1
            self._fillscreen()
            self._need_redraw = True
            return True

        return False

    def pagedown(self):
        if self.height != 1:
            curpos = self.pos
            self.vert_align(self.portto - 1, top=True)
            ret = self.pos != curpos
            if ret:
                self._need_redraw = True
            return ret
        else:
            return self.linedown()

    def pageup(self):
        if self.height != 1:
            curpos = self.pos
            self.vert_align(self.portfrom, bottom=True)
            ret = self.pos != curpos
            if ret:
                self._need_redraw = True
            return ret
        else:
            return self.lineup()



import functools


def command(cmdid):
    def _f(f):
        f.COMMAND_ID = cmdid
        return f
    return _f


def is_enable(cmdid):
    def _f(f):
        f.COMMAND_IS_ENABLE = cmdid
        return f
    return _f


def norec(f):
    f.NOREC = True
    return f


def norerun(f):
    f.NORERUN = True
    return f

# def repeat(f):
#    def rep(self, wnd, *args, **kwargs):
#        n = 1 if not wnd.editmode.has_repeat() else wnd.editmode.get_repeat()
#        for i in range(n):
#            f(self, wnd, *args, **kwargs)
#    functools.update_wrapper(rep, f)
#    return rep


class Commands:

    def _find_funcs(self, attrname):
        for name in dir(self):
            f = getattr(self, name)
            cmdid = getattr(f, attrname, None)
            if cmdid:
                yield (cmdid, f)

    def get_commands(self):
        yield from self._find_funcs('COMMAND_ID')

    def get_commands_is_enable(self):
        yield from self._find_funcs('COMMAND_IS_ENABLE')



import kaa
from kaa.keyboard import *
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style
from kaa.filetype.default import modebase, keybind
from kaa.ui.msgbox import msgboxmode
from kaa.command import command, Commands, norec, norerun
from kaa.commands import editorcommand
from kaa.ui.selectlist import filterlist
from gappedbuf.sre_constants import error as gre_error

SearchThemes = {
    'basic': [],
}


class SearchCommands(editorcommand.EditCommands):
    # todo: move commands to mode class

    @command('searchdlg.search.next')
    @norec
    @norerun
    def search_next(self, wnd):
        mode = wnd.document.mode
        mode.search_next(wnd)

    @command('searchdlg.search.prev')
    @norec
    @norerun
    def search_prev(self, wnd):
        mode = wnd.document.mode
        mode.search_prev(wnd)

    def _show_histdlg(self, wnd, title, candidates, callback):
        filterlist.show_listdlg(title, candidates, callback)

    @command('searchdlg.history')
    @norec
    @norerun
    def search_history(self, wnd):
        def callback(result):
            if result:
                f, t = wnd.document.marks['searchtext']
                wnd.document.replace(f, t, result)
                wnd.cursor.setpos(f)

        self._show_histdlg(wnd, 'Recent searches',
                           [s for s,
                               info in kaa.app.config.hist(
                                   'search_text').get()],
                           callback)

    @command('searchdlg.toggle.ignorecase')
    @norec
    @norerun
    def toggle_ignorecase(self, wnd):
        mode = wnd.document.mode
        mode.toggle_option_ignorecase()

    @command('searchdlg.toggle.wordsearch')
    @norec
    @norerun
    def toggle_wordsearch(self, wnd):
        mode = wnd.document.mode
        mode.toggle_option_wordsearch()

    @command('searchdlg.toggle.regex')
    @norec
    @norerun
    def toggle_regex(self, wnd):
        mode = wnd.document.mode
        mode.toggle_option_regex()

searchdlg_keys = {
    '\r': ('searchdlg.search.next'),
    '\n': ('searchdlg.search.next'),
    up: ('searchdlg.history'),
}


class SearchDlgMode(dialogmode.DialogMode):
    autoshrink = True

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
    ]
    _last_searchstr = ''

    def __init__(self, target):
        super().__init__()

        self.target = target
        self.initialpos = target.cursor.pos
        self.initialrange = target.screen.selection.get_selrange()
        self.lastsearch = None
        self.lasthit = None
        self.lastmatch = None
        self.option = modebase.SearchOption.LAST_SEARCH

        if target and target.screen.selection.is_selected():
            f, t = target.screen.selection.get_selrange()
            s = target.document.gettext(f, t).split('\n')
            if s:
                s = s[0].strip()
                if s:
                    self.option.text = s

    def close(self):
        super().close()
        self.target = None
        kaa.app.messagebar.set_message("")

    def init_keybind(self):
        super().init_keybind()

        self.register_keys(self.keybind, self.KEY_BINDS)
        self.keybind.add_keybind(searchdlg_keys)

    def init_commands(self):
        super().init_commands()

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

        self._searchcommands = SearchCommands()
        self.register_command(self._searchcommands)

    def init_themes(self):
        super().init_themes()
        self.themes.append(SearchThemes)

    def create_cursor(self, wnd):
        return dialogmode.DialogCursor(wnd,
                                       [dialogmode.MarkRange('searchtext')])

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        wnd.set_cursor(self.create_cursor(wnd))
        if self.option.text:
            wnd.screen.selection.set_range(*self.document.marks['searchtext'])
        wnd.cursor.setpos(self.document.marks['searchtext'][1])

    def _build_input(self, f):
        # search text
        f.append_text('caption', 'Search:')
        f.append_text('default', ' ')
        f.append_text('default', '', mark_pair='searchtext')
        f.append_text('default', ' ')

    def _build_buttons(self, f):
        # buttons
        f.append_text('checkbox', '[&Next]',
                      shortcut_style='button.shortcut',
                      on_shortcut=self.search_next)

        f.append_text('checkbox', '[&Prev]',
                      shortcut_style='button.shortcut',
                      on_shortcut=self.search_prev)

    def _build_options(self, f):
        f.append_text('checkbox', '[&Ignore case]',
                      mark_pair='ignore-case',
                      on_shortcut=self.toggle_option_ignorecase,
                      shortcut_style='checkbox.shortcut',
                      shortcut_mark='shortcut-i')

        f.append_text('checkbox', '[&Word]',
                      mark_pair='word',
                      on_shortcut=self.toggle_option_word,
                      shortcut_style='checkbox.shortcut',
                      shortcut_mark='shortcut-w')

        f.append_text('checkbox', '[&Regex]',
                      mark_pair='regex',
                      on_shortcut=self.toggle_option_regex,
                      shortcut_style='checkbox.shortcut',
                      shortcut_mark='shortcut-r')

    def build_document(self):
        with dialogmode.FormBuilder(self.document) as f:
            self._build_input(f)
            self._build_buttons(f)
            self._build_options(f)

            self.document.insert(
                self.document.marks['searchtext'][0], self.option.text)

            self.update_option_style()

        kaa.app.messagebar.set_message(
            "Hit alt+N/alt+P to search Next/Prev. Hit up to show history.")

    def _set_option_style(self, mark, style,
                          shortcutmark, shortcutstyle):
        f, t = self.document.marks[mark]
        self.document.setstyles(f, t, self.get_styleid(style))

        f = self.document.marks[shortcutmark]
        self.document.setstyles(f, f + 1, self.get_styleid(shortcutstyle))

    def _get_optionstylename(self, f):
        if f:
            return '.checked'
        else:
            return ''

    def update_option_style(self):
        style = self._get_optionstylename(self.option.ignorecase)
        self._set_option_style('ignore-case', 'checkbox' + style, 'shortcut-i',
                               'checkbox.shortcut' + style)

        style = self._get_optionstylename(self.option.word)
        self._set_option_style('word', 'checkbox' + style, 'shortcut-w',
                               'checkbox.shortcut' + style)

        style = self._get_optionstylename(self.option.regex)
        self._set_option_style('regex', 'checkbox' + style, 'shortcut-r',
                               'checkbox.shortcut' + style)

    def _option_updated(self):
        self.update_option_style()
#        self.document.style_updated()

    def toggle_option_ignorecase(self, wnd):
        self.option.ignorecase = not self.option.ignorecase
        self._option_updated()

    def toggle_option_word(self, wnd):
        self.option.word = not self.option.word
        self._option_updated()

    def toggle_option_regex(self, wnd):
        self.option.regex = not self.option.regex
        self._option_updated()

    def get_search_str(self):
        f, t = self.document.marks['searchtext']
        return self.document.gettext(f, t)

    def on_edited(self, wnd):
        self.lastsearch = None
        self.lasthit = None
        self.lastmatch = None
        s = self.get_search_str()

        try:
            if s != self._last_searchstr:
                if not self._search_next(wnd):
                    self.lastsearch = -1
                    self._search_next(wnd)
            self._last_searchstr = s
        except gre_error as e:
            kaa.app.messagebar.set_message(str(e))

    def _show_searchresult(self, hit):
        if hit:
            f, t = hit.span()
            self.target.cursor.setpos(f)
            self.target.screen.selection.set_range(f, t)
            kaa.app.messagebar.set_message('found')
        else:
            self.target.screen.selection.clear()
            kaa.app.messagebar.set_message('not found')

    def _search_next(self, wnd):
        self.option.text = self.get_search_str()
        if not self.option.text:
            self.target.screen.selection.clear()
        else:
            if self.lasthit is self.lastsearch is None:
                # initial search
                if self.initialrange:
                    pos = self.initialrange[0]
                else:
                    pos = self.initialpos

            elif self.lasthit:
                # continue previous search
                pos = self.lasthit[0] + 1

            else:
                # resume from top
                pos = 0

            ret = self.target.document.mode.search_next(pos, self.option)
            self._show_searchresult(ret)

            self.lastsearch = pos
            self.lasthit = ret.span() if ret else None
            self.lastmatch = ret

            return ret

    def _save_searchstr(self):
        s = self.get_search_str()
        if s:
            kaa.app.config.hist('search_text').add(s)

    def search_next(self, wnd):
        self._save_searchstr()
        return self._search_next(wnd)

    def _search_prev(self, wnd):
        self.option.text = self.get_search_str()
        if self.option.text:
            if self.lasthit is self.lastsearch is None:
                # initial search
                if self.initialrange:
                    pos = self.initialrange[1]
                else:
                    pos = self.initialpos

            elif self.lasthit:
                # continue previous search
                pos = self.lasthit[1] - 1

            else:
                # resume from end
                pos = self.target.document.endpos()

            ret = self.target.document.mode.search_prev(pos, self.option)
            self._show_searchresult(ret)

            self.lastsearch = pos
            self.lasthit = ret.span() if ret else None
            self.lastmatch = ret

            return ret

    def search_prev(self, wnd):
        self._save_searchstr()
        return self._search_prev(wnd)

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        self._save_searchstr()
        self.target.activate()
        self.target = None
        wnd.get_label('popup').destroy()


class ReplaceCommands(SearchCommands):
    # todo: move commands to mode class

    @command('replacedlg.field.next')
    @norec
    @norerun
    def field_next(self, wnd):
        searchfrom, searchto = wnd.document.marks['searchtext']
        replacefrom, replaceto = wnd.document.marks['replacetext']

        if searchfrom <= wnd.cursor.pos <= searchto:
            wnd.cursor.setpos(replacefrom)
            wnd.screen.selection.set_range(replacefrom, replaceto)
        else:
            wnd.cursor.setpos(searchfrom)
            wnd.screen.selection.set_range(searchfrom, searchto)

    def replace_repl_history(self, wnd):
        def callback(result):
            if result:
                f, t = wnd.document.marks['replacetext']
                wnd.document.replace(f, t, result)
                wnd.cursor.setpos(f)

        self._show_histdlg(wnd, 'Recent replace strings',
                           [s for s,
                               info in kaa.app.config.hist('repl_text').get()],
                           callback)

    @command('replacedlg.history')
    @norec
    @norerun
    def replace_history(self, wnd):
        searchfrom, searchto = wnd.document.marks['searchtext']

        if searchfrom <= wnd.cursor.pos <= searchto:
            self.search_history(wnd)
        else:
            self.replace_repl_history(wnd)


replacedlg_keys = {
    '\r': ('replacedlg.field.next'),
    '\n': ('replacedlg.field.next'),
    up: ('replacedlg.history'),
}


class ReplaceDlgMode(SearchDlgMode):
    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
    ]

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(replacedlg_keys)

    def init_commands(self):
        super().init_commands()
        self._replacecommands = ReplaceCommands()
        self.register_command(self._replacecommands)

    def create_cursor(self, wnd):
        return dialogmode.DialogCursor(wnd,
                                       [dialogmode.MarkRange('searchtext'),
                                        dialogmode.MarkRange('replacetext')])

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

    def _build_input(self, f):
        # search text
        f.append_text('caption', 'Search: ')
        f.append_text('default', ' ')
        f.append_text('default', '', mark_pair='searchtext')
        f.append_text('default', '\n')

        # replace text
        f.append_text('caption', 'Replace:')
        f.append_text('default', ' ')
        f.append_text('default', '', None, mark_pair='replacetext')
        f.append_text('default', '\n')

    def _build_buttons(self, f):
        # buttons
        super()._build_buttons(f)
        f.append_text('checkbox', '[&All]', 'start-all', 'end-all',
                      shortcut_style='button.shortcut', on_shortcut=self.replace_all)

    def build_document(self):
        with dialogmode.FormBuilder(self.document) as f:
            self._build_input(f)
            self._build_buttons(f)
            self._build_options(f)

            self.document.insert(
                self.document.marks['searchtext'][0], self.option.text)

            self.document.insert(
                self.document.marks['replacetext'][0], self.option.replace_to)

            self.update_option_style()

        kaa.app.messagebar.set_message(
            "Hit enter to move field. Hit up to show history.`")
        return

    def get_replace_str(self):
        f, t = self.document.marks['replacetext']
        return self.document.gettext(f, t)

    def _show_replace_msg(self, wnd, on_y, on_n):
        def cb(c):
            wnd.activate()
            if c == 'y':
                on_y(wnd)
            elif c == 'n':
                on_n(wnd)
            elif c == 'a':
                self.replace_all(wnd)
            else:
                pass

        msgdoc = msgboxmode.MsgBoxMode.show_msgbox(
            'Replace text?', ['&Yes', '&No', '&All', '&Cancel'], cb)

    def _show_search_again(self, wnd, on_y):
        def cb(c):
            wnd.activate()
            if c == 'r':
                on_y(wnd)

        msgdoc = msgboxmode.MsgBoxMode.show_msgbox(
            'Search failed. Resume again?', ['&Resume', '&Cancel'], cb)

    def _save_replstr(self):
        kaa.app.config.hist('repl_text').add(self.get_replace_str())

    def search_next(self, wnd):
        if not self.get_search_str():
            return

        self.option.replace_to = self.get_replace_str()
        self._save_searchstr()
        self._save_replstr()

        if self.lastsearch is None:
            self._search_next(wnd)

        if self.lasthit:
            self._show_replace_msg(
                wnd, self.replace_and_next, self.skip_and_next)
        else:
            self._show_search_again(wnd, self.skip_and_next)

    _metachars = [
        (r"\\", "\\"),
        (r"\a", "\a"),
        (r"\b", "\b"),
        (r"\f", "\f"),
        (r"\n", "\n"),
        (r"\r", "\r"),
        (r"\t", "\t"),
    ]

    def _repl_str(self, update_cursor):
        f, t = self.lasthit
        if self.option.regex:
            s = self.lastmatch.expand(self.option.replace_to)
        else:
            s = self.option.replace_to

        self.target.document.mode.replace_string(
            self.target, f, t, s, update_cursor=update_cursor)

        self.lasthit = (f, f + len(s))

    def replace_and_next(self, wnd):
        self._repl_str(update_cursor=True)
        self._search_next(wnd)
        self.search_next(wnd)

    def skip_and_next(self, wnd):
        self._search_next(wnd)
        self.search_next(wnd)

    def search_prev(self, wnd):
        if not self.get_search_str():
            return

        self.option.replace_to = self.get_replace_str()
        self._save_searchstr()
        self._save_replstr()

        self._search_prev(wnd)

        if self.lasthit:
            self._show_replace_msg(
                wnd, self.replace_and_prev, self.skip_and_prev)
        else:
            self._show_search_again(wnd, self.skip_and_prev)

    def replace_and_prev(self, wnd):
        self._repl_str(update_cursor=True)
        self.search_prev(wnd)

    def skip_and_prev(self, wnd):
        self.search_prev(wnd)

    def replace_all(self, wnd):
        self.option.replace_to = self.get_replace_str()
        self.option.text = self.get_search_str()

        self._save_searchstr()
        self._save_replstr()

        if not self.option.text:
            return
        self.lastsearch = None
        newstr = self.get_replace_str()

        if self.target.document.undo:
            self.target.document.undo.beginblock()
        pos = 0
        n = 0
        lastpos = None
        try:
            while True:
                ret = self.target.document.mode.search_next(pos, self.option)
                self.lasthit = ret.span() if ret else None
                self.lastmatch = ret

                if ret:
                    self._repl_str(update_cursor=False)
                    pos = lastpos = self.lasthit[1]
                    n += 1
                else:
                    break
        finally:
            if self.target.document.undo:
                self.target.document.undo.endblock()

        if lastpos is not None:
            self.target.cursor.setpos(lastpos)

        kaa.app.messagebar.set_message('Replaced {} time(s)'.format(n))

    def on_esc_pressed(self, wnd, event):
        self._save_replstr()
        super().on_esc_pressed(wnd, event)



from kaa.filetype import filetypedef

class FileTypeInfo(filetypedef.FileTypeInfo):
    FILE_EXT = {'.diff', '.patch'}

    @classmethod
    def get_modetype(cls):
        from kaa.filetype.diff.diffmode import DiffMode
        return DiffMode



import sys
import traceback
from contextlib import contextmanager
import kaa
from kaa import document
from kaa.keyboard import *
from kaa.filetype.default import defaultmode
from kaa.filetype.python import pythonmode
from kaa.command import command, Commands, norec, norerun
from kaa.theme import Theme, Style

pythonconsole_keys = {
    (alt, '\r'): ('python.exec'),
    (alt, '\n'): ('python.exec'),
}

PythonConsoleThemes = {
    'basic': [
        Style('stdout', 'Default', 'Default'),
        Style('stderr', 'Magenta', 'Default'),
    ],
}


class PythonConsoleMode(pythonmode.PythonMode):
    MODENAME = 'Python console'
    DOCUMENT_MODE = False

    PYTHONCONSOLE_KEY_BINDS = [
        pythonconsole_keys,
    ]

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, self.PYTHONCONSOLE_KEY_BINDS)

    @contextmanager
    def _redirect_output(self):
        stdout = sys.stdout
        stderr = sys.stderr

        output = self.output
        style_stdout = output.mode.get_styleid('stdout')
        style_stderr = output.mode.get_styleid('stderr')

        class out:

            def __init__(self, style):
                self.style = style

            def write(self, s):
                if not output.closed:
                    output.append(s, self.style)

        sys.stdout = out(style_stdout)
        sys.stderr = out(style_stderr)

        yield

        sys.stdout = stdout
        sys.stderr = stderr

        w = output.wnds[0]
        if not w.closed:
            # rebuild screen before move cursor
            w.screen.locate(w.screen.pos, top=True)
            w.cursor.eof()

    def _eval_str(self, s):
        try:
            comp = compile(s, '<interactive input>', 'eval')
        except SyntaxError:
            return

        try:
            import __main__
            ret = eval(comp, __main__.__dict__, __main__.__dict__)
            if ret is not None:
                print(repr(ret))
        except Exception:
            traceback.print_exc()
            return True

        return True

    def _exec_str(self, s):
        try:
            comp = compile(s, '<interactive input>', 'exec')
        except SyntaxError:
            traceback.print_exc()
            return

        try:
            import __main__
            exec(comp, __main__.__dict__, __main__.__dict__)
        except Exception:
            traceback.print_exc()
            return True

        return True

    @command('python.exec')
    @norec
    @norerun
    def exec_script(self, wnd):
        is_sel = wnd.screen.selection.is_selected()
        if is_sel:
            f, t = wnd.screen.selection.get_selrange()
            s = wnd.document.gettext(f, t)
        else:
            s = wnd.document.gettext(0, wnd.document.endpos())

        s = s.rstrip()
        if not s:
            return

        scr = s + '\n'

        with self._redirect_output():
            if not self._eval_str(s):
                self._exec_str(s)

        if not is_sel:
            endpos = self.document.endpos()
            if not self.document.gettext(endpos - 1, endpos).endswith('\n'):
                self.insert_string(wnd, self.document.endpos(),
                                                 '\n', update_cursor=False)
            wnd.screen.selection.set_range(0, endpos)


class PythonOutputMode(defaultmode.DefaultMode):
    MODENAME = 'Python outputs'
    DOCUMENT_MODE = False
    USE_UNDO = False
    NO_WRAPINDENT = True

    def init_themes(self):
        super().init_themes()
        self.themes.append(PythonConsoleThemes)


def show_console():
    buf = document.Buffer()
    cons = document.Document(buf)
    mode = PythonConsoleMode()
    cons.setmode(mode)
    cons.set_title('<Python console>')

    buf = document.Buffer()
    output = document.Document(buf)
    mode = PythonOutputMode()
    output.setmode(mode)
    output.set_title('<Python output>')

    cons.mode.output = output

    frame = kaa.app.show_doc(cons).get_label('frame')
    frame.splitter.split(vert=False, doc=output)

    kaa.app.messagebar.set_message('Hit alt+Enter to execute script.')



import kaa
from kaa import document
from kaa.ui.filenameindex import filenameindexmode
from gappedbuf import re as gre


def show(commandline, s):
    buf = document.Buffer()
    doc = document.Document(buf)
    doc.set_title(' '.join(commandline.split()))
    mode = MakeoutputMode()
    doc.setmode(mode)

    doc.insert(0, s)

    mode = doc.mode
    style_filename = mode.get_styleid('filenameindex-filename')
    style_lineno = mode.get_styleid('filenameindex-lineno')

    for m in mode.RE_FILENAME.finditer(doc.buf):
        f, t = m.span('FILENAME')
        doc.setstyles(f, t, style_filename, update=False)

        f, t = m.span('LINENO')
        doc.setstyles(f, t, style_lineno, update=False)

    kaa.app.show_doc(doc)


class MakeoutputMode(filenameindexmode.FilenameIndexMode):
    MODENAME = 'Make'



class Context:

    def iter_roots(self):
        c = self
        while c:
            if isinstance(c, ContextRoot):
                yield c
            c = c.get_context_parent()

    def get_context_parent(self):
        raise NotImplementedError

    def get_context(self):
        for r in self.iter_roots():
            return r

    def set_label(self, name, obj):
        self.get_context().context[name] = obj

    def get_label(self, name):
        for r in self.iter_roots():
            ret = r.context.get(name)
            if ret:
                return ret

    def del_label(self, name):
        if name in self.get_context().context:
            del self.context[name]
            return True


class ContextRoot(Context):

    def __init__(self):
        super().__init__()
        self.context = {}

    def destroy_context(self):
        self.context.clear()



import re
import kaa
from kaa import document, command
from kaa.theme import Theme, Style
from kaa.filetype.default import keybind
from kaa.commands import editorcommand
from kaa.keyboard import *
from kaa.ui.dialog import dialogmode
from kaa.ui.selectlist import filterlist


def number_filter(wnd, s):
    if s == '0':
        t = wnd.document.mode.get_input_text().strip()
        if not t or int(t) == 0:
            return ''
    return re.match(r'\d*', s).group()


InputlineThemes = {
    'basic': []
}

inputline_keys = {
    ('\r'): ('inputline'),
    ('\n'): ('inputline'),
    up: ('inputline.history'),
}


class InputlineMode(dialogmode.DialogMode):
    autoshrink = True
    USE_UNDO = True
    auto_indent = False

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
        inputline_keys
    ]

    def init_themes(self):
        super().init_themes()
        self.themes.append(InputlineThemes)

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, self.KEY_BINDS)

    def init_commands(self):
        super().init_commands()

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        wnd.set_cursor(dialogmode.DialogCursor(wnd,
                                               [dialogmode.MarkRange('inputtext')]))

        f, t = self.document.marks['inputtext']
        wnd.cursor.setpos(f)

        if f != t:
            wnd.screen.selection.set_range(f, t)

    def on_esc_pressed(self, wnd, event):
        # todo: run callback
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("Canceled")

    def filter_string(self, wnd, s):
        if self.filter:
            s = self.filter(wnd, s)
        return s

    def get_input_text(self):
        f, t = self.document.marks['inputtext']
        s = self.document.gettext(f, t)
        return s

    def set_input_ext(self, wnd, s):
        cur = self.get_input_text()
        f, t = self.document.marks['inputtext']
        self.document.replace(f, t, s, self.get_styleid('default'))
        f, t = self.document.marks['inputtext']
        wnd.cursor.setpos(f)
        if s:
            wnd.screen.selection.set_range(f, t)

    @command.command('inputline')
    @command.norec
    @command.norerun
    def input_line(self, w):
        s = self.get_input_text()
        self.callback(w, s)
        popup = w.get_label('popup')
        popup.destroy()

    @command.command('inputline.history')
    @command.norec
    @command.norerun
    def input_history(self, wnd):
        if not self.history:
            return

        def callback(result):
            if result:
                self.set_input_ext(wnd, result)

        filterlist.show_listdlg(self.caption, self.history, callback)

    @classmethod
    def build(cls, caption, callback, filter=None, history=(), value=''):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)
        mode.caption = caption
        mode.callback = callback
        mode.filter = filter
        mode.history = history

        with dialogmode.FormBuilder(doc) as f:
            # caption
            f.append_text('caption', caption)
            f.append_text('default', ' ')
            f.append_text('default', value, mark_pair='inputtext')

        return doc



# key definition

import itertools
import kaa
from kaa import log
from kaa.exceptions import KaaError


class KeyBind:
    translator = None

    def __init__(self):
        self.keylist = []

    def clear(self):
        self.keylist = []

    def _split_modifier(self, key):
        mods = set()
        ch = None
        for k in key:
            if isinstance(k, Modifier):
                if k in mods:
                    raise KaaError(
                        'Multiple modifiers: {!r}'.format(key))
                mods.add(k)
            else:
                if ch is not None:
                    raise KaaError(
                        'Multiple characters: {!r}'.format(key))

                if isinstance(k, str):
                    if len(k) != 1:
                        raise KaaError(
                            'Invalid characters: {!r}'.format(key))

                ch = k

        return (mods, ch)

    def _flatten_keys(self, keys):
        if isinstance(keys, str):
            return keys

        if isinstance(keys, SpecialKey):
            return [[keys]]

        if [k for k in keys if isinstance(k, Modifier)]:
            return [keys]

        ret = []
        for k in keys:
            if isinstance(k, (list, tuple, set)):
                ret.append(k)
            elif isinstance(k, str):
                ret.extend(k)
            else:
                ret.append([k])
        return ret

    def add_keybind(self, d):
        for keys, commands in d.items():
            if isinstance(commands, str):
                commands = [commands]
            try:
                keys = self._flatten_keys(keys)
                keys = [self._split_modifier(k) for k in keys]
                keys = list(itertools.chain.from_iterable(
                    kaa.app.translate_key(mod, c) for mod, c in keys))
                self.keylist.insert(0, (keys, commands))

            except KaaError as e:
                log.warning('Failed to load key: {}'.format(e))

    def get_command(self, keys):
        for cmdkeys, commands in self.get_candidates(keys):
            if cmdkeys == keys:
                return commands

    def get_candidates(self, keys):
        for cmdkeys, commands in self.keylist:
            if keys == cmdkeys[:len(keys)]:
                yield cmdkeys, commands


class Modifier:
    keys = {}

    def __init__(self, name):
        self.name = name
        self.keys[name] = self

    def __repr__(self):
        return '<Modifier:{}>'.format(self.name)


class SpecialKey:
    keys = {}

    def __init__(self, name):
        self.name = name
        self.keys[name] = self

    def __repr__(self):
        return '<Key:{}>'.format(self.name)


shift = Modifier('shift')
ctrl = Modifier('ctrl')
alt = Modifier('alt')

backspace = SpecialKey('backspace')
pageup = SpecialKey('pageup')
pagedown = SpecialKey('pagedown')
left = SpecialKey('left')
right = SpecialKey('right')
up = SpecialKey('up')
down = SpecialKey('down')
insert = SpecialKey('insert')
delete = SpecialKey('delete')
home = SpecialKey('home')
end = SpecialKey('end')
esc = SpecialKey('esc')
tab = SpecialKey('tab')
f1 = SpecialKey('f1')
f2 = SpecialKey('f2')
f3 = SpecialKey('f3')
f4 = SpecialKey('f4')
f5 = SpecialKey('f5')
f6 = SpecialKey('f6')
f7 = SpecialKey('f7')
f8 = SpecialKey('f8')
f9 = SpecialKey('f9')
f10 = SpecialKey('f10')
f11 = SpecialKey('f11')
f12 = SpecialKey('f12')
f13 = SpecialKey('f13')
f14 = SpecialKey('f14')
f15 = SpecialKey('f15')
f16 = SpecialKey('f16')
f17 = SpecialKey('f17')
f18 = SpecialKey('f18')
f19 = SpecialKey('f19')
f20 = SpecialKey('f20')
f21 = SpecialKey('f21')
f22 = SpecialKey('f22')
f23 = SpecialKey('f23')
f24 = SpecialKey('f24')



import kaa
from kaa.command import Commands, command, is_enable, norec, norerun
from kaa import document

# todo: following imports should be defered
from kaa.ui.mainmenu import menumode
from kaa.ui.itemlist import itemlistmode
from kaa.ui.moveseparator import moveseparatormode


class ApplicationCommands(Commands):

    @command('app.mainmenu')
    @norec
    @norerun
    def show_mainmenu(self, wnd):
        menumode.MenuMode.show_menu(wnd, 'MAIN')

    @command('menu.window')
    @norec
    @norerun
    def show_windowmenu(self, wnd):
        menumode.MenuMode.show_menu(wnd, 'CHANGE-WINDOW')

    @command('menu.edit.convert')
    @norec
    @norerun
    def show_editmenuconvert(self, wnd):
        menumode.MenuMode.show_menu(wnd, 'EDIT-CONVERT')

    @command('menu.code')
    @norec
    @norerun
    def show_codemenu(self, wnd):
        menumode.MenuMode.show_menu(wnd, 'CODE')

    def _walk_all_wnds(self, wnd):
        yield wnd
        curframe = wnd.get_label('frame')
        if curframe:
            for w in curframe.get_editors():
                if w is not wnd:
                    yield w

        for frame in kaa.app.get_frames():
            if frame is not curframe:
                for w in frame.get_editors():
                    yield w

    @command('app.global.prev')
    @norec
    @norerun
    def global_prev(self, wnd):
        for w in self._walk_all_wnds(wnd):
            if w.document.mode.on_global_prev(w):
                return

    @command('app.global.next')
    @norec
    @norerun
    def global_next(self, wnd):
        for w in self._walk_all_wnds(wnd):
            if w.document.mode.on_global_next(w):
                return

    # todo: move following methods somewhere else

    @command('app.show-framelist')
    @norec
    @norerun
    def show_framelist(self, wnd):
        from kaa.ui.framelist import framelistmode
        doc = framelistmode.FrameListMode.build()
        kaa.app.show_dialog(doc)

    @command('editor.splitvert')
    @norec
    @norerun
    def editor_splitvert(self, wnd):
        if wnd.splitter:
            s = wnd.splitter.split(vert=True)
            s.wnd.activate()

    @command('editor.splithorz')
    @norec
    @norerun
    def editor_splithorz(self, wnd):
        if wnd.splitter:
            s = wnd.splitter.split(vert=False)
            s.wnd.activate()

    @command('editor.moveseparator')
    @norec
    @norerun
    def editor_moveseparator(self, wnd):
        moveseparatormode.move_separator(wnd)

    @command('editor.nextwindow')
    @norec
    def editor_nextwindow(self, wnd):
        splitter = wnd.parent.splitter
        if splitter:
            wnds = [wnd for s, wnd in splitter.walk() if wnd]
            try:
                n = wnds.index(wnd) + 1
            except ValueError:
                return

            if n >= len(wnds):
                n = 0
            wnds[n].activate()

    @command('editor.prevwindow')
    @norec
    def editor_prevwindow(self, wnd):
        splitter = wnd.parent.splitter
        if splitter:
            wnds = [wnd for s, wnd in splitter.walk() if wnd]
            try:
                n = wnds.index(wnd) - 1
            except ValueError:
                return

            if n < 0:
                n = len(wnds) - 1
            wnds[n].activate()

    def save_splitterdocs(self, wnd, splitter, callback):
        wnds = set()
        docs = set()
        for child, w in splitter.walk():
            if w:
                wnds.add(w)
                docs.add(w.document)

        saves = [doc for doc in docs if wnds.issuperset(doc.wnds)]

        wnd.document.mode.file_commands.save_documents(wnd, saves, callback)

    @command('editor.joinwindow')
    @norec
    def editor_joinwindow(self, wnd):
        if wnd.splitter and wnd.splitter.parent:
            buddy = wnd.splitter.get_buddy()
            if not buddy:
                return   # not split

            def saved(canceled):
                if not canceled:
                    wnd.splitter.parent.join(wnd)

            self.save_splitterdocs(wnd, buddy, saved)

    @command('editor.switchfile')
    @norec
    @norerun
    def editor_switchfile(self, wnd):
        docs = []
        for frame in kaa.app.get_frames():
            docs.extend(frame.get_documents())

        titles = [doc.get_title() for doc in docs]

        curdoc = wnd.document
        curdoc_close = curdoc.mode.CLOSE_ON_DEL_WINDOW

        def callback(n):
            curdoc.mode.CLOSE_ON_DEL_WINDOW = curdoc_close
            if n is not None:
                doc = docs[n]
                wnd.show_doc(doc)
            else:
                wnd.show_doc(curdoc)

        def selchanged(n):
            doc = docs[n]
            wnd.show_doc(doc)

        def saved(canceled):
            if canceled:
                return

            curdoc.mode.CLOSE_ON_DEL_WINDOW = False
            n = docs.index(wnd.document)
            doc = itemlistmode.ItemListMode.build(
                '', titles, n, callback, selchanged)
            kaa.app.show_dialog(doc)

        wnd.document.mode.file_commands.can_close_wnd(wnd, saved)



import os, sys
from setuptools import setup, find_packages, Extension


if sys.version_info[:2] < (3, 3):
    raise RuntimeError('Kaa requires Python 3.3 or later.')

try:
    from Cython.Distutils import build_ext
    cmdclass = {'build_ext': build_ext}

    ext = Extension("_gappedbuf", 
        ["_gappedbuf/_gappedbuf.pyx",
         "_gappedbufre/_sre.c"])

except ImportError:
    cmdclass = {}
    ext = Extension("_gappedbuf", 
        ["_gappedbuf/_gappedbuf.c",
         "_gappedbufre/_sre.c"])

def read(fname):
    return open(
            os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    cmdclass = cmdclass,
    name = "kaaedit",
    version = "0.21.0",
    description='kaa - console text editor.',
    url='https://github.com/kaaedit/kaa',
    author='Atsuo Ishimoto',
    author_email='ishimoto@gembook.org',
    long_description=read('README.rst'),
    classifiers=[
            "Programming Language :: Python :: 3",
            "Development Status :: 3 - Alpha",
            "Topic :: Text Editors",
            "Topic :: Software Development :: Debuggers",
            "Environment :: Console :: Curses",
            "License :: OSI Approved :: MIT License", ],
    license='MIT License',
    install_requires=['curses_ex', 'pyjf3', 'setproctitle', 'kaadbg'],
    packages = find_packages(),
    ext_modules = [ext],
    entry_points = {
        'console_scripts': [
            'kaa = kaa.cui.main:run',
        ],}
)




import kaa_testutils
from kaa.commands import editorcommand


class TestCursor(kaa_testutils._TestScreenBase):

    def test_cursor(self):
        wnd = self._getwnd("abc\ndef\n")

        cmd = editorcommand.CursorCommands()
        cmd.right(wnd)
        assert wnd.cursor.pos == 1

        cmd.left(wnd)
        assert wnd.cursor.pos == 0

        cmd.down(wnd)
        assert wnd.cursor.pos == 4

        cmd.up(wnd)
        assert wnd.cursor.pos == 0


class TestEdit(kaa_testutils._TestScreenBase):

    def test_putstring(self):
        wnd = self._getwnd("abc\ndef\n")

        wnd.document.mode.put_string(wnd, '12345')
        assert wnd.document.gettext(0, 5) == '12345'

    def test_delete(self):
        wnd = self._getwnd("abc\ndef\n")

        cmd = editorcommand.EditCommands()
        cmd.delete(wnd)
        assert wnd.document.gettext(0, 2) == 'bc'

        wnd = self._getwnd("a\u0301bc\ndef\n")

        cmd = editorcommand.EditCommands()
        cmd.delete(wnd)
        assert wnd.document.gettext(0, 2) == 'bc'

    def test_backspace(self):
        wnd = self._getwnd("abc\ndef\n")

        wnd.cursor.setpos(1)
        cmd = editorcommand.EditCommands()
        cmd.backspace(wnd)
        assert wnd.document.gettext(0, 2) == 'bc'
        assert wnd.cursor.pos == 0

        wnd = self._getwnd("a\u0302bc\ndef\n")

        wnd.cursor.setpos(2)
        cmd = editorcommand.EditCommands()
        cmd.backspace(wnd)
        assert wnd.document.gettext(0, 2) == 'bc'
        assert wnd.cursor.pos == 0

    def test_lineindent(self):
        wnd = self._getwnd("abc")
        wnd.cursor.setpos(0)
        cmd = editorcommand.EditCommands()
        cmd.indent(wnd)
        assert wnd.document.gettext(0, 7) == '    abc'

    def test_blockindent(self):
        wnd = self._getwnd("abc\ndef\nghi")
        wnd.screen.selection.set_range(1, 9)
        cmd = editorcommand.EditCommands()
        cmd.indent(wnd)
        assert wnd.document.gettext(0, 7) == '    abc'

    def test_blockindentblanklines(self):
        wnd = self._getwnd("abc\n\ndef method")
        wnd.screen.selection.set_range(1, 6)
        cmd = editorcommand.EditCommands()
        cmd.indent(wnd)
        assert wnd.document.gettext(0, 16) == "    abc\n\n    def"

        wnd = self._getwnd("a\na")
        wnd.screen.selection.set_range(0, 3)
        cmd.indent(wnd)
        assert wnd.document.gettext(0, 12) == "    a\n    a"
        
    def test_undo_ins(self):
        wnd = self._getwnd("")
        cmd = editorcommand.EditCommands()

        wnd.document.mode.put_string(wnd, 'abc')

        cmd.undo(wnd)
        assert wnd.document.endpos() == 0

        cmd.redo(wnd)
        assert wnd.document.gettext(0, 3) == 'abc'

    def test_undo_delete(self):
        wnd = self._getwnd("abc")
        cmd = editorcommand.EditCommands()

        cmd.delete(wnd)

        cmd.undo(wnd)
        assert wnd.document.gettext(0, 3) == 'abc'

        cmd.redo(wnd)
        assert wnd.document.gettext(0, 2) == 'bc'

    def test_undo_backspace(self):
        wnd = self._getwnd("abc")
        cmd = editorcommand.EditCommands()

        wnd.cursor.setpos(1)
        cmd.backspace(wnd)

        cmd.undo(wnd)
        assert wnd.document.gettext(0, 3) == 'abc'
        assert wnd.cursor.pos == 1

        cmd.redo(wnd)
        assert wnd.document.gettext(0, 2) == 'bc'
        assert wnd.cursor.pos == 0

    def test_undo_newline(self):
        wnd = self._getwnd("    abc")
        cmd = editorcommand.EditCommands()

        wnd.cursor.setpos(7)
        cmd.newline(wnd)

        assert wnd.document.gettext(0, 7) == '    abc'
        assert wnd.document.gettext(7, 12) == '\n    '

        wnd = self._getwnd("      abc")
        cmd = editorcommand.EditCommands()

        wnd.cursor.setpos(4)
        cmd.newline(wnd)

        assert wnd.document.gettext(0, 5) == '\n    '
        assert wnd.document.gettext(5, 15) == '  abc'



from libc.stdlib cimport malloc, free, realloc
from libc.string cimport memmove, memcpy
from cpython cimport unicode, mem, ref
from cpython.list cimport PyList_New, PyList_SET_ITEM
from cpython.long cimport PyLong_FromUnsignedLong

cimport rtdef

cdef extern from "_gappedbufre.h":
    object PyInit__gappedbufre()



def init_gappedbufre():
    import sys
    sys.modules['_gappedbufre'] = PyInit__gappedbufre()

cdef class GappedBuffer:
    def __cinit__(self):
        DEF BUFSIZE = 4096

        self.buf = <Py_UCS4*>malloc(BUFSIZE * sizeof(Py_UCS4))
        if not self.buf:
            raise MemoryError()

        self.bufsize = BUFSIZE
        self.numelems = 0
        self.gap = 0
        self.gapsize = BUFSIZE

    def __dealloc__(self):
        if self.buf:
            free(self.buf)
            self.buf = NULL
    
    cdef void _resize_buf(self, Py_ssize_t size):
        cdef Py_UCS4 *buf = <Py_UCS4*>realloc(
                self.buf, size * sizeof(Py_UCS4))

        if not buf:
            raise MemoryError()
        self.buf = buf
        self.bufsize = size

    cdef void _expand_gap(self, Py_ssize_t size):
        if size <= self.gapsize:
            return
        self._resize_buf(self.numelems+size)

        if self.gap < self.numelems:
            # resize gap
            memmove(self.buf+self.gap+size, self.buf+self.gap+self.gapsize,
                    (self.numelems-self.gap) * sizeof(Py_UCS4))
    
        self.gapsize = size

    cdef void _shrink_gap(self):
        DEF MAX_GAPSIZE = 4096
        if self.numelems + MAX_GAPSIZE > self.bufsize:
            return
        memmove(self.buf+self.gap+MAX_GAPSIZE,
                self.buf+self.gap+self.gapsize,
                (self.numelems-self.gap) * sizeof(Py_UCS4))

        self.gapsize = MAX_GAPSIZE
        self._resize_buf(self.numelems+MAX_GAPSIZE)

    cdef void _move_gap(self, Py_ssize_t index):
        assert index <= self.bufsize
        if self.gap == index:
            return
        if self.gapsize:
            if index < self.gap:
                memmove(self.buf+index+self.gapsize, self.buf+index,
                        (self.gap - index)*sizeof(Py_UCS4))
            else:
                memmove(self.buf+self.gap, self.buf+self.gap+self.gapsize,
                        (index - self.gap)*sizeof(Py_UCS4))
        self.gap = index
    
    cdef _insert(self, Py_ssize_t index, unicode s):
        cdef Py_ssize_t size = len(s)
        if not (0 <= index <= self.numelems):
            raise ValueError('Invalid index value: %d, %d' % (
                                index, self.numelems))

        if size:
            self._move_gap(index)
            if self.gapsize < size:
                self._expand_gap(max(1024, size*2))
            rtdef.PyUnicode_AsUCS4(s, self.buf+index, size, False)
            
            self.numelems += size
            self.gap += size
            self.gapsize -= size

    cpdef insert(self, Py_ssize_t index, unicode s):
        self._insert(index, s)

    cpdef append(self, unicode s):
        self.insert(self.numelems, s)

    cdef _delete(self, Py_ssize_t begin, Py_ssize_t end):
        if not (0 <= begin <= end <= self.numelems):
            raise ValueError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))
        
        cdef Py_ssize_t size = end-begin
        if not size:
            return

        self._move_gap(begin)
        self.gapsize += size
        self.numelems -= size
        self._shrink_gap()

    cpdef delete(self, Py_ssize_t begin, Py_ssize_t end):
        self._delete(begin, end)

    cpdef replace(self, Py_ssize_t begin, Py_ssize_t end, unicode s):
        self._delete(begin, end)
        self._insert(begin, s)

    cdef Py_UCS4 _get_max(self, Py_ssize_t begin, Py_ssize_t end) nogil:
        cdef Py_UCS4 maxchar = 0
        cdef Py_ssize_t b, e

        e = min(end, self.gap)

        cdef Py_ssize_t p
        for p from begin <= p < e:
            maxchar = max(maxchar, self.buf[p])
        b = max(begin, self.gap) 
        for p from b <= p < end:
            maxchar = max(maxchar, self.buf[p+self.gapsize])
        return maxchar

    cdef void _copy_chars1(self, void *dest, Py_ssize_t begin, 
            Py_ssize_t end) nogil:
        cdef rtdef.Py_UCS1 *buf = <rtdef.Py_UCS1 *>dest
        cdef Py_ssize_t b, e

        # copy before gap
        e = min(end, self.gap)
        cdef Py_ssize_t p
        for p from begin <= p < e:
            buf[0] = self.buf[p]
            buf += 1

        # copy after gap
        b = max(begin, self.gap) + self.gapsize
        for p from b <= p < end+self.gapsize:
            buf[0] = self.buf[p]
            buf += 1
        
    cdef void _copy_chars2(self, void *dest, Py_ssize_t begin, 
            Py_ssize_t end) nogil:
        cdef rtdef.Py_UCS2 *buf = <rtdef.Py_UCS2 *>dest
        cdef Py_ssize_t b, e

        # copy before gap
        e = min(end, self.gap)
        cdef Py_ssize_t p
        for p from begin <= p < e:
            buf[0] = self.buf[p]
            buf += 1

        # copy after gap
        b = max(begin, self.gap) + self.gapsize
        for p from b <= p < end+self.gapsize:
            buf[0] = self.buf[p]
            buf += 1

    cdef void _copy_chars4(self, void *dest, Py_ssize_t begin, 
            Py_ssize_t end) nogil:
        cdef Py_ssize_t n = end - begin
        cdef Py_ssize_t former

        if self.gap < begin:
            memcpy(dest, self.buf+begin+self.gapsize, n*sizeof(Py_UCS4))
        elif end <= self.gap:
            memcpy(dest, self.buf+begin, n*sizeof(Py_UCS4))
        else:
            former = self.gap - begin
            memcpy(dest, self.buf+begin, former*sizeof(Py_UCS4))
            memcpy(dest+former*sizeof(Py_UCS4), self.buf+self.gap+self.gapsize, 
                    (n-former)*sizeof(Py_UCS4))

    cpdef get(self, Py_ssize_t begin, Py_ssize_t end):
        if not (0 <= begin <= end <= self.numelems):
            raise IndexError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))

        cdef Py_ssize_t n = end - begin
        cdef Py_UCS4 maxchar = self._get_max(begin, end)
        cdef unicode ret = rtdef.PyUnicode_New(n, maxchar)
        cdef void *buf = rtdef.PyUnicode_DATA(ret)
        cdef nbytes = rtdef.PyUnicode_KIND(ret)

        if nbytes == 1:
            self._copy_chars1(buf, begin, end)
        elif nbytes == 2:
            self._copy_chars2(buf, begin, end)
        elif nbytes == 4:
            self._copy_chars4(buf, begin, end)
        else:
            raise RuntimeError()

        return ret
    
    def __len__(self):
        return self.numelems
    
    cdef int _get_slice(self, object obj, 
            Py_ssize_t *f, Py_ssize_t *t) except *:
        cdef Py_ssize_t step, slicelength

        if rtdef.PySlice_Check(obj):
            rtdef.PySlice_GetIndicesEx(obj, self.numelems, f, t, 
                     &step, &slicelength)
            if step != 1:
                raise ValueError('step is not supported')
            return 1
        else: 
            f[0] = obj
            if f[0] < 0:
                f[0] += self.numelems
            if not (0 <= f[0] < self.numelems):
                raise IndexError('Invalid index value: %d, %d' % (
                                    f[0], self.numelems))
            return 0

    def __getitem__(self, object item):
        cdef Py_ssize_t start, stop
        
        if self._get_slice(item, &start, &stop):
            return self.get(start, stop)
        else:
            return self.buf[start+(0 if start < self.gap else self.gapsize)]

    def __setitem__(self, object item, unicode value):
        cdef Py_ssize_t start, stop
        
        if self._get_slice(item, &start, &stop):
            self.replace(start, stop, value)
        else:
            if len(value) != 1:
                raise ValueError('Invalid string length: %d' % len(value))
            
            start = start+(0 if start < self.gap else self.gapsize)
            self.buf[start] = <Py_UCS4>value

    def __delitem__(self, object item):
        cdef Py_ssize_t start, stop
        
        if self._get_slice(item, &start, &stop):
            self.delete(start, stop)
        else: 
            self.delete(start, start+1)

    def findchr(self, unicode c, Py_ssize_t begin, Py_ssize_t end):
        cdef Py_UCS4 *chars
        cdef Py_ssize_t size, b, e, p

        if not (0 <= begin <= end <= self.numelems):
            raise ValueError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))
        
        chars = rtdef.PyUnicode_AsUCS4Copy(c)
        try:
            size = len(c)
            # search before gap
            e = min(end, self.gap)
            for p from begin <= p < e:
                for q from 0 <= q < size:
                    if self.buf[p] == chars[q]:
                        return p

            # search after gap
            b = max(begin, self.gap) + self.gapsize
            for p from b <= p < end+self.gapsize:
                for q from 0 <= q < size:
                    if self.buf[p] == chars[q]:
                        return p-self.gapsize
        finally:
            mem.PyMem_Free(chars)

        return -1

    def rfindchr(self, unicode c, Py_ssize_t begin, Py_ssize_t end):
        cdef Py_UCS4 *chars
        cdef Py_ssize_t size, b, e, p

        if not (0 <= begin <= end <= self.numelems):
            raise ValueError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))
        
        chars = rtdef.PyUnicode_AsUCS4Copy(c)
        try:
            size = len(c)
            # search after gap
            b = max(begin, self.gap) + self.gapsize
            for p from (end+self.gapsize) > p >= b:
                for q from 0 <= q < size:
                    if self.buf[p] == chars[q]:
                        return p-self.gapsize

            # search before gap
            e = min(end, self.gap)
            for p from e > p >= begin:
                for q from 0 <= q < size:
                    if self.buf[p] == chars[q]:
                        return p
        finally:
            mem.PyMem_Free(chars)

        return -1
    
    cpdef getint(self, Py_ssize_t pos):
        if not (0 <= pos < self.numelems):
            raise ValueError('Invalid index value: %d, %d' % (
                                pos, self.numelems))
        if pos < self.gap:
            return <unsigned long>self.buf[pos]
        else:
            return <unsigned long>self.buf[self.gapsize+pos]

    cpdef getints(self, Py_ssize_t begin, Py_ssize_t end):

        cdef Py_ssize_t b, e, p, n
        cdef list ret
        cdef object v

        if not (0 <= begin <= end <= self.numelems):
            raise ValueError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))
        
        ret = PyList_New(end-begin)
        n = 0 
        # read before gap
        e = min(end, self.gap)
        for p from begin <= p < e:
            v = <unsigned long>self.buf[p]
            PyList_SET_ITEM(ret, n, v)
            ref.Py_INCREF(v)
            n += 1

        # read after gap
        b = max(begin, self.gap) + self.gapsize
        for p from b <= p < end+self.gapsize:
            v = <unsigned long>self.buf[p]
            PyList_SET_ITEM(ret, n, v)
            ref.Py_INCREF(v)
            n += 1

        return ret

    cdef _insertints(self, Py_ssize_t index, object s):

        cdef Py_ssize_t size = len(s)
        if not (0 <= index <= self.numelems):
            raise ValueError('Invalid index value: %d, %d' % (
                                index, self.numelems))

        if size:
            self._move_gap(index)
            if self.gapsize < size:
                self._expand_gap(max(1024, size*2))

            for value in s:
                self.buf[self.gap] = <unsigned long>value
                self.gap += 1

            self.numelems += size
            self.gapsize -= size

    cpdef insertints(self, Py_ssize_t index, object s):

        self._insertints(index, s)

    cpdef appendints(self, object s):
        self.insertints(self.numelems, s)

    cpdef replaceints(self, Py_ssize_t begin, Py_ssize_t end, object s):
        self._delete(begin, end)
        self._insertints(begin, s)

    cpdef setints(self, Py_ssize_t begin, Py_ssize_t end, unsigned long v):
        cdef Py_ssize_t b, e, p

        if not (0 <= begin <= end <= self.numelems):
            raise ValueError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))

        n = 0
        # replace before gap
        e = min(end, self.gap)
        for p from begin <= p < e:
            self.buf[p] = v

        # replace after gap
        b = max(begin, self.gap) + self.gapsize
        for p from b <= p < end+self.gapsize:
            self.buf[p] = v

    def findint(self, object nums, Py_ssize_t begin, Py_ssize_t end,
                int comp_ne):
        cdef unsigned long *longs
        cdef Py_ssize_t size, b, e, p

        if not (0 <= begin <= end <= self.numelems):
            raise ValueError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))

        size = len(nums)
        longs = <unsigned long *>mem.PyMem_Malloc(
                                    size * sizeof(unsigned long))
        if not longs:
            raise MemoryError()

        try:
            for i from 0 <= i < size:
                longs[i] = nums[i]

            # search before gap
            e = min(end, self.gap)
            for p from begin <= p < e:
                if not comp_ne:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            return p
                else:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            break
                    else:
                        return p

            # search after gap
            b = max(begin, self.gap) + self.gapsize
            for p from b <= p < end+self.gapsize:
                if not comp_ne:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            return p-self.gapsize
                else:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            break
                    else:
                        return p-self.gapsize
        finally:
            mem.PyMem_Free(longs)

        return -1

    def rfindint(self, object nums, Py_ssize_t begin, Py_ssize_t end,
                 int comp_ne):

        cdef unsigned long *longs
        cdef Py_ssize_t size, b, e, p

        if not (0 <= begin <= end <= self.numelems):
            raise ValueError('Invalid range: %d, %d, %d' % (
                                begin, end, self.numelems))

        size = len(nums)
        longs = <unsigned long *>mem.PyMem_Malloc(
                                    size * sizeof(unsigned long))
        if not longs:
            raise MemoryError()
        try:
            for i from 0 <= i < size:
                longs[i] = nums[i]

            # search after gap
            b = max(begin, self.gap) + self.gapsize
            for p from (end+self.gapsize) > p >= b:
                if not comp_ne:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            return p-self.gapsize
                else:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            break
                    else:
                        return p-self.gapsize

            # search before gap
            e = min(end, self.gap)
            for p from e > p >= begin:
                if not comp_ne:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            return p
                else:
                    for q from 0 <= q < size:
                        if self.buf[p] == longs[q]:
                            break
                    else:
                        return p
        finally:
            mem.PyMem_Free(longs)

        return -1



import re
import kaa
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.ui.selectlist import selectlist

from kaa.theme import Theme, Style

from kaa.command import command, norec, norerun
from kaa.keyboard import *
from kaa.commands import editorcommand
import kaa.filetype.default.keybind


FilterListThemes = {
    'basic': [],
}


class FilterListMode(selectlist.SelectItemList):
    SEP = '\n'
    MAX_CAPTION_LEN = None
    USE_PHRASE_STYLE = False

    def init_themes(self):
        super().init_themes()
        self.themes.append(FilterListThemes)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)
        wnd.set_label('filterlist', wnd)

    def set_candidates(self, candidates):
        ws = re.compile(r'\s')
        self.candidates = []
        for c in candidates:
            phrase = False
            caption = c
            if ws.search(caption):
                phrase = True
                caption = ' '.join(caption.split())
                if not caption:
                    caption = '(...)'

            if self.MAX_CAPTION_LEN is not None:
                if len(caption) > self.MAX_CAPTION_LEN:
                    caption = caption[:self.MAX_CAPTION_LEN] + '...'
                    phrase = True

            if not self.USE_PHRASE_STYLE or not phrase:
                c = selectlist.SelectItem(
                    'selectitem', 'selectitem-active', caption, c)
            else:
                c = selectlist.SelectItem(
                    'selectphrase', 'selectphrase-active', caption, c)
            self.candidates.append(c)

    def _filter_items(self, query):
        query = [f.upper() for f in query.split()]
        if query:
            items = []
            for item in self.candidates:
                u = item.text.upper()
                for q in query:
                    if q not in u:
                        break
                else:
                    items.append(item)
        else:
            items = self.candidates[:]

        sel = items[0] if items else None
        return sel, items

    def set_query(self, wnd, query):
        sel, items = self._filter_items(query)
        self.update_doc(items)
        if sel:
            self.update_sel(wnd, sel)


FilterListInputDlgThemes = {
    'basic': [],
}


filterlistinputdlg_keys = {
    down: 'filterlistdlg.next',
    (ctrl, 'n'): 'filterlistdlg.next',
    up: 'filterlistdlg.prev',
    (ctrl, 'p'): 'filterlistdlg.prev',
    '\r': 'filterlistdlg.select',
    '\n': 'filterlistdlg.select',
}


class FilterListInputDlgMode(dialogmode.DialogMode):
    MAX_INPUT_HEIGHT = 4
    callback = None
    INITIAL_MESSAGE = "Hit up/down to select item."
    NO_WRAPINDENT = False

    @classmethod
    def build(cls, caption, callback):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)

        mode.callback = callback

        with dialogmode.FormBuilder(doc) as f:
            f.append_text('caption', caption)
            f.append_text('default', ' ')
            f.append_text('default', '', mark_pair='query')

        return doc

    def init_themes(self):
        super().init_themes()
        self.themes.append(FilterListInputDlgThemes)

    def init_keybind(self):
        self.keybind.add_keybind(
            kaa.filetype.default.keybind.edit_command_keys)
        self.keybind.add_keybind(kaa.filetype.default.keybind.cursor_keys)
        self.keybind.add_keybind(kaa.filetype.default.keybind.emacs_keys)
        self.keybind.add_keybind(filterlistinputdlg_keys)

    def init_commands(self):
        super().init_commands()

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        wnd.CURSOR_TO_MIDDLE_ON_SCROLL = False
        cursor = dialogmode.DialogCursor(wnd,
                                         [dialogmode.MarkRange('query')])

        wnd.set_cursor(cursor)
        wnd.cursor.setpos(self.document.marks['query'][1])
        wnd.set_label('query_field', self)
        kaa.app.messagebar.set_message(self.INITIAL_MESSAGE)

    def calc_position(self, wnd):
        w, h = wnd.getsize()
        height = self.calc_height(wnd)
        height = min(height, self.MAX_INPUT_HEIGHT)
        top = wnd.mainframe.height - height - wnd.mainframe.MESSAGEBAR_HEIGHT
        return 0, top, wnd.mainframe.width, top + height

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("")
        if self.callback:
            self.callback(None)

    def on_edited(self, wnd):
        filterlist = wnd.get_label('filterlist')
        filterlist.document.mode.set_query(wnd, self.get_query())
        filterlist.get_label('popup').on_console_resized()

    def get_query(self):
        f, t = self.document.marks['query']
        return self.document.gettext(f, t)

    def set_query(self, wnd, s):
        wnd.screen.selection.clear()
        f, t = self.document.marks['query']
        self.document.replace(f, t, s)
        wnd.cursor.setpos(f + len(s))

    @command('filterlistdlg.next')
    @norec
    @norerun
    def next(self, wnd):
        filterlist = wnd.get_label('filterlist')
        filterlist.document.mode.sel_next(filterlist)

    @command('filterlistdlg.prev')
    @norec
    @norerun
    def prev(self, wnd):
        filterlist = wnd.get_label('filterlist')
        filterlist.document.mode.sel_prev(filterlist)

    def _selected(self, wnd, value):
        popup = wnd.get_label('popup')
        popup.destroy()

        if self.callback:
            self.callback(value)

    @command('filterlistdlg.select')
    @norec
    @norerun
    def on_select(self, wnd):
        filterlist = wnd.get_label('filterlist')
        cur = filterlist.document.mode.cursel
        if cur:
            self._selected(wnd, cur.value)
            return True


def show_listdlg(title, candidates, callback):
    doc = FilterListInputDlgMode.build(
        title, callback)
    dlg = kaa.app.show_dialog(doc)

    filterlistdoc = FilterListMode.build()
    list = dlg.add_doc('dlg_filterlist', 0, filterlistdoc)

    filterlistdoc.mode.set_candidates(candidates)
    filterlistdoc.mode.set_query(list, '')
    dlg.on_console_resized()

    return dlg



import kaa_testutils
from kaa.commands import editorcommand
import kaa.macro
from kaa.command import command, norec


class TestMacro(kaa_testutils._TestScreenBase):

    def test_toggle_record(self):
        macro = kaa.macro.Macro()
        assert not macro.is_recording()

        macro.toggle_record()
        assert macro.is_recording()

        macro.toggle_record()
        assert not macro.is_recording()

    def test_record(self):

        macro = kaa.macro.Macro()
        macro.toggle_record()

        @command('command1')
        def f():
            pass

        macro.record(f)
        assert macro.get_commands() == [('command1', (), {})]

        @command('command2')
        @norec
        def g():
            pass

        macro.record(g)
        assert macro.get_commands() == [('command1', (), {})]

    def test_run(self):
        wnd = self._getwnd("abc\ndef\n")

        macro = kaa.macro.Macro()
        macro.toggle_record()

        macro.record(wnd.document.mode.cursor_commands.right)
        macro.record(wnd.document.mode.cursor_commands.right)
        macro.record(wnd.document.mode.cursor_commands.right)
        macro.record(wnd.document.mode.cursor_commands.left)
        macro.toggle_record()

        macro.run(wnd)

        assert wnd.cursor.pos == 2

        macro.toggle_record()

        macro.record_string('123')
        macro.record_string('45')
        macro.toggle_record()

        macro.run(wnd)

        assert wnd.cursor.pos == 7
        assert wnd.document.gettext(0, 7) == 'ab12345'



from contextlib import ExitStack
from unittest.mock import patch

import kaa
from kaa.ui.framelist import framelistmode
import kaa_testutils


class _frame:

    def get_title(self):
        return 'title'


class TestFrameList(kaa_testutils._TestDocBase):

    def test_framelist(self):
        with ExitStack() as st:
            frames = [_frame(), _frame()]
            st.enter_context(patch('kaa.app.get_frames',
                                   create=True, return_value=frames))
            st.enter_context(patch('kaa.app.get_activeframe',
                                   create=True, return_value=frames[0]))

            doc = framelistmode.FrameListMode.build()



from unittest.mock import patch
from kaa import keyboard
from kaa.keyboard import *


class TestKeyBind:

    def _create_keybind(self, bind):

        kb = keyboard.KeyBind()
        kb.add_keybind(bind)

        return kb

    def test_char(self):
        kb = self._create_keybind({
            'a': 'command'
        })

        assert kb.keylist == [
            ([set(), 'a'], ['command']),
        ]

        kb = self._create_keybind({
            'aa': 'command'
        })

        assert kb.keylist == [
            ([set(), 'a', set(), 'a'], ['command']),
        ]

    def test_special(self):
        kb = self._create_keybind({
            right: 'command'
        })

        assert kb.keylist == [
            ([set(), right], ['command']),
        ]

        kb = self._create_keybind({
            (right, left): 'command'
        })

        assert kb.keylist == [
            ([set(), right, set(), left], ['command']),
        ]

        kb = self._create_keybind({
            ((right,), (left,)): 'command'
        })

        assert kb.keylist == [
            ([set(), right, set(), left, ], ['command']),
        ]

    def test_mix(self):

        kb = self._create_keybind({
            ((right,), 'a'): 'command'
        })

        assert kb.keylist == [
            ([set(), right, set(), 'a'], ['command']),
        ]

        kb = self._create_keybind({
            (right, 'a'): 'command'
        })

        assert kb.keylist == [
            ([set(), right, set(), 'a'], ['command']),
        ]

        kb = self._create_keybind({
            ((right,), 'ab', left): 'command'
        })

        assert kb.keylist == [
            ([set(), right, set(), 'a', set(), 'b', set(), left], ['command']),
        ]

    def test_modifier(self):
        kb = self._create_keybind({
            ('a', ctrl): 'command'
        })

        assert kb.keylist == [
            ([set([ctrl]), 'a'], ['command']),
        ]

        kb = self._create_keybind({
            (('a', ctrl), ('b', ctrl)): 'command'
        })

        assert kb.keylist == [
            ([set([ctrl]), 'a', set([ctrl]), 'b'], ['command']),
        ]

        kb = self._create_keybind({
            (ctrl, shift, right): 'command'
        })

        assert kb.keylist == [
            ([set([ctrl, shift]), right], ['command']),
        ]

        kb = self._create_keybind({
            ((ctrl, right), (ctrl, 'b')): 'command'
        })

        assert kb.keylist == [
            ([set([ctrl]), right, set([ctrl]), 'b'], ['command']),
        ]

    def test_callable(self):

        def f(wnd):
            pass

        kb = self._create_keybind({
            ((right,), 'a'): f
        })

        assert kb.keylist == [
            ([set(), right, set(), 'a'], f),
        ]



import difflib
import kaa
from kaa import document
from kaa.filetype.diff import diffmode
from kaa.filetype.default import keybind
from kaa.ui.dialog import dialogmode
from kaa.commands import editorcommand


class ViewDiffMode(dialogmode.DialogMode):
    MODENAME = 'ViewDiff'
    DOCUMENT_MODE = False
    USE_UNDO = False

    def init_commands(self):
        super().init_commands()

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
    ]

    def init_keybind(self):
        super().init_keybind()

        self.register_keys(self.keybind, self.KEY_BINDS)

    def init_themes(self):
        super().init_themes()
        self.themes.append(diffmode.DiffThemes)

    def init_tokenizers(self):
        self.tokenizers = [diffmode.build_tokenizer()]

    def is_cursor_visible(self):
        return 1   # hide cursor

    def calc_position(self, wnd):
        height = wnd.screen.get_total_height()
        height = min(height, wnd.mainframe.height * 3 // 4)
        top = wnd.mainframe.height - height - wnd.mainframe.MESSAGEBAR_HEIGHT
        return 0, top, wnd.mainframe.width, top + height

    def on_str(self, wnd, s):
        # does nothing
        pass

    def on_esc_pressed(self, wnd, event):
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("")
        if self.callback:
            self.callback()


def view_diff(curdoc, callback=None):
    orig = kaa.app.storage.openfile(
        curdoc.fileinfo.fullpathname,
        curdoc.fileinfo.encoding,
        curdoc.fileinfo.newline,
        nohist=True)

    cur_lines = list(curdoc.iterlines(0))
    org_lines = list(orig.iterlines(0))

    diff = ''.join(difflib.unified_diff(org_lines, cur_lines,
                                        curdoc.fileinfo.fullpathname, '(buffer)'))

    buf = document.Buffer()
    doc = document.Document(buf)
    doc.insert(0, diff)

    mode = ViewDiffMode()
    mode.callback = callback
    doc.setmode(mode)

    kaa.app.show_dialog(doc)



import kaa
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.ui.selectlist import filterlist
from kaa.command import command, norec, norerun
from kaa.keyboard import *


class WordCompleteList(filterlist.FilterListMode):
    MAX_CAPTION_LEN = 30
    USE_PHRASE_STYLE = True

    def _filter_items(self, query):
        if query:
            query = query.upper()
            items = []
            for s in self.candidates:
                u = s.text.upper()
                if u.startswith(query):
                    items.append(s)
        else:
            items = self.candidates[:]

        sel = items[0] if items else None
        return sel, items


workcomplete_keys = {
    tab: 'filterlistdlg.next',
    (shift, tab): 'filterlistdlg.prev',
    '\r': 'wordcomplete.select',
    '\n': 'wordcomplete.select',
}


class WordCompleteInputMode(filterlist.FilterListInputDlgMode):
    MAX_INPUT_HEIGHT = 4
    INITIAL_MESSAGE = "Hit tab/shift+tab to complete."

    @classmethod
    def build(cls, target):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)
        mode.target = target

        with dialogmode.FormBuilder(doc) as f:
            f.append_text('caption', 'Select word:')
            f.append_text('default', ' ')
            f.append_text('default', '', mark_pair='query')

        return doc

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(workcomplete_keys)

    def calc_position(self, wnd):
        w, h = wnd.getsize()
        height = self.calc_height(wnd)
        height = min(height, self.MAX_INPUT_HEIGHT)

        cury, curx = self.orgloc
        if self.stack_upper:
            top = max(0, cury - height)
        else:
            top = cury + 1

        return 0, top, wnd.mainframe.width, top + height

    @command('wordcomplete.select')
    @norec
    @norerun
    def selected(self, wnd):
        self.target.screen.selection.clear()
        list = wnd.get_label('dlg_filterlist')
        sel = list.document.mode.cursel
        if sel:
            s = sel.value
        else:
            s = self.get_query()

        self.target.document.mode.replace_string(
            self.target, self.wordpos[0], self.wordpos[1],
            s, update_cursor=True)
        wnd.get_label('popup').destroy()

    def on_edited(self, wnd):
        super().on_edited(wnd)
        s = self.get_query()
        f, t = self.wordpos
        self.target.document.mode.replace_string(
            self.target, f, t, s, update_cursor=True)
        self.wordpos = (f, f + len(s))

    def on_esc_pressed(self, wnd, event):
        self.target.screen.selection.clear()
        super().on_esc_pressed(wnd, event)

    def start(self, list):
        self.orgpos = self.target.cursor.pos

        self.wordpos = (self.orgpos, self.orgpos)
        wnd = self.document.wnds[0]

        curword = ''
        word = self.target.document.mode.get_word_at(self.orgpos)
        if word:
            f, t, cg = word
            t = max(f, min(t, self.orgpos))
            if f < t and cg[0] in 'LMN':  # Letter, Mark, Number
                self.wordpos = (f, t)

                curword = self.target.document.gettext(f, t)
                if curword:
                    self.target.screen.selection.set_range(f, t)
                    self.set_query(wnd, curword)

        # build word list
        # word at cursor position should not appear in the list.
        words = [w for w in self.target.document.mode.get_word_list()
                 if w != curword]

        words.extend(kaa.app.clipboard.get_all())

        list.document.mode.set_candidates(words)
        list.document.mode.candidates.sort(key=lambda v: v.text.upper())
        list.document.mode.set_query(list, self.get_query())
        list.get_label('popup').on_console_resized()


def show_wordlist(wnd):
    doc = WordCompleteInputMode.build(wnd)
    doc.mode.orgloc = wnd.get_cursor_loc()

    if doc.mode.orgloc[0] > (wnd.mainframe.height // 2):
        doc.mode.stack_upper = True
    else:
        doc.mode.stack_upper = False

    dlg = kaa.app.show_dialog(doc)

    filterlistdoc = WordCompleteList.build()
    filterlistdoc.mode.SEP = ' '
    list = dlg.add_doc('dlg_filterlist', 0, filterlistdoc)

    doc.mode.start(list)

    return dlg
